"""
# Calculates RMSDs for a selection.

## Algorithm

Calculates the RMSD values along a trajectory slice for different
selections. If multiple selections are given creates a series data
for that selection.

RMSD is calculated using [libcalc.mda_rmsd](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.mda_rmsd).

## Examples

Calculate RMSD of the whole system:

    taurenmd rmsd top.pdb traj.dcd -e rmsd.csv

Calculates RMSDs for different selections:

    taurenmd rmsd top.pdb traj.dcd -g 'segid A' 'segid B' -e

``-x`` exports the data to a CSV file. You can also plot the data with
the ``-v`` option:

    [...] -x rmsd.csv -v title=my-plot-title xlabel=frames ylabel=RMSDs ...

where ``[...]`` is the previous command example.

You can also use ``tmdrmsd`` instead of ``taurenmd rmsd``.

## References

"""  # noqa: E501
import argparse
import functools
from datetime import datetime

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log
from taurenmd.libs import libcalc, libcli, libio, libmda, libplot
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_plottemplates_param}'
    )

_help = 'Calculates RMSDs for a selection and trajectory slice.'
_name = 'rmsd'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_atom_selections_arg(ap)
libcli.add_reference_frame_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        start=None,
        stop=None,
        step=None,
        ref_frame=0,
        selections=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs
        ):
    """Execute main client logic."""
    log.info(T('starting'))
    
    u = libmda.load_universe(topology, *trajectories)
    
    frame_slice = libio.frame_slice(
        start=start,
        stop=stop,
        step=step,
        )
    
    if selections is None:
        selections = ['all']

    rmsds = []
    for selection in selections:
        rmsds.append(
            libcalc.mda_rmsd(
                u,
                frame_slice=frame_slice,
                selection=selection,
                ref_frame=ref_frame,
                )
            )
    if export:
        libio.export_data_to_file(
            list(range(len(u.trajectory))[frame_slice]),
            *rmsds,
            fname=export,
            delimiter=',',
            header=(
                "# Date: {}\n'"
                "# Topology: {}\n"
                "# Trajectories : {}\n"
                "# ref frame: {}\n"
                "# frame number,{}\n"
                ).format(
                    datetime.now(),
                    Path(topology).resolve(),
                    ', '.join(f.resolve().str() for f in trajectories),
                    ref_frame,
                    ','.join(selections),
                    ),
            )
    
    if plot:
        plotvars = plotvars or dict()
        plotvars.setdefault('labels', selections)

        log.info(T('plot params:'))
        for k, v in plotvars.items():
            log.info(S('{} = {!r}', k, v))
        
        libplot.param(
            list(range(len(u.trajectory))[frame_slice]),
            rmsds,
            **plotvars,
            )

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""General vars for tests."""

from taurenmd.core import Path

datap = Path(
    Path(__file__).myparents(),
    'data',
    )

export_data_expected = Path(datap, 'export_data_expected.csv')
export_data_expected_2 = Path(datap, 'export_data_expected_2.csv')
toptest = Path(datap, 'pcnaA_frame0.pdb')
toptest_cif = Path(datap, 'pcnaA_frame0.cif')
trajtest = Path(datap, 'pcnaA.dcd')



# -*- coding: utf-8 -*-
"""Sphinx config file."""
from __future__ import unicode_literals

import os
import mock
import sys


mock_modules = [
    'MDAnalysis',
    'MDAnalysis.analysis',
    'MDAnalysis.analysis.rms',
    'mdtraj',
    'simtk.openmm.app',
    'simtk.openmm',
    'simtk',
    'bioplottemplates',
    'bioplottemplates.plots',
    'pyquaternion',
    ]

for modulename in mock_modules:
    sys.modules[modulename] = mock.Mock()

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinxarg.ext',
    'sphinx.ext.autosectionlabel',
    ]

source_suffix = '.rst'
master_doc = 'index'
project = 'taurenmd'
year = '2019'
author = 'Joao MC Teixeira'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.8.10'

todo_include_todos = True
pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/joaomcteixeira/taurenmd/issues/%s', '#'),
    'pr': ('https://github.com/joaomcteixeira/taurenmd/pull/%s', 'PR #'),
    'mda': ('https://www.mdanalysis.org', 'MDAnalysis'),
    'mdtraj': ('http://mdtraj.org/1.9.3', 'MDTraj'),
    'openmm': ('https://openmmtools.readthedocs.io/en/0.18.1', 'OpenMM'),
    }
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

linkcheck_ignore = [r'https://codecov.io/*']

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = 'sphinx_rtd_theme'

html_logo = 'logo/taurenmd_logo_black.png'
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
    '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
    }
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



"""
Shared operations for client interfaces.

This module contains functions and classes that are shared amongst the
client interfaces. It contains also others used to enhance the user
experience.
"""
import argparse
import sys
from datetime import datetime
from functools import wraps

import taurenmd.core as tcore
from taurenmd import _BANNER, references
from taurenmd.logger import CMDFILE


def load_args(ap):
    """Load user arguments."""
    cmd = ap.parse_args()
    return cmd


def maincli(ap, main):
    """
    Client main function.

    Operates when client is called directly outside the
    ``taurenmd`` client interface.
    
    - Reads input parameters
    - saves inpu command to log file
    - runs client ``main`` function
    - saves references to log file
    
    Returns
    -------
    The result value from client ``main`` function.
    """
    cmd = load_args(ap)
    save_command(CMDFILE, *sys.argv)
    result = main(**vars(cmd))
    save_references()
    return result


def add_reference(ref):
    """
    Add reference decorator.

    Example
    -------

        >>> @add_reference(str)
        >>> def myfunct():
        >>>     ...
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            references.add(ref)
            result = func(*args, **kwargs)
            return result
        return wrapper
    return decorator


def save_references():
    """Save used references to log file."""
    out = [f'{i}: {ref}' for i, ref in enumerate(sorted(references), start=2)]
    with open(CMDFILE, 'a') as fh:
        fh.write('References:\n')
        fh.write(f'1: {tcore.ref_taurenmd}')
        fh.write(''.join(out))


# https://stackoverflow.com/questions/4042452
class CustomParser(argparse.ArgumentParser):
    """Custom Parser class."""

    def error(self, message):
        """Present error message."""
        self.print_help()
        self.print_usage()
        sys.stderr.write('*** INPUT ERROR: {}\n'.format(message))
        sys.stderr.write('*** read the usage details above\n')
        sys.exit(2)


class ParamsToDict(argparse.Action):
    """
    Convert command-line parameters in an argument to a dictionary.

    Example
    -------
    
    Where ``-x`` is an optional argument of the command-line client
    interface.

        >>> par1=1 par2='my name' par3=[1,2,3]
        >>> {'par1': 1, 'par2': 'my name', 'par3': [1, 2, 3]}

    """

    def __call__(self, parser, namespace, values, option_string=None):
        """Execute."""
        bool_value = {
            'true': True,
            'false': False,
            }

        param_dict = {}
        for kv in values:
            # print(param_dict, kv)
            try:
                k, v = kv.split('=')
            except ValueError:
                param_dict[kv] = True
            else:
                if ',' in v:
                    vs = v.split(',')
                    try:
                        param_dict[k] = tuple(float(i) for i in vs)
                    except (ValueError, TypeError):
                        param_dict[k] = tuple(i for i in vs)

                else:
                    try:
                        param_dict[k] = float(v)
                    except (ValueError, TypeError):  # is string or list
                        param_dict[k] = bool_value.get(v.lower(), v)
            
        namespace.plotvars = param_dict
        setattr(namespace, self.dest, True)


def save_command(fname, *args):
    """
    Append the execution command to a log file.

    Parameters
    ----------
    fname : string or Path
        The file name of the log file where to append the command.

    *args : strings
        String parts that compose the command.
    """
    with open(fname, 'a') as fh:
        fh.write(
            '[{}] {}\n'.format(
                datetime.now().strftime("%d/%m/%Y, %H:%M:%S"),
                ' '.join(str(a) for a in args),
                )
            )


def add_subparser(parser, module):
    """
    Add a subcommand to a parser.

    Parameters
    ----------
    parser : `argparse.add_suparsers object <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_subparsers>`_
        The parser to add the subcommand to.

    module
        A python module containing the characteristics of a taurenmd
        client interface. Client interface modules require the following
        attributes: ``__doc__`` which feeds the `description argument <https://docs.python.org/3/library/argparse.html#description>`_
        of `add_parser <https://docs.python.org/3/library/argparse.html#other-utilities>`_,
        ``_help`` which feeds `help <https://docs.python.org/3/library/argparse.html#help>`_,
        ``ap`` which is an `ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_,
        and a ``main`` function, which executes the main logic of the interface.
    """  # noqa: E501
    new_ap = parser.add_parser(
        module._name,
        description=module.ap.description,
        help=module._help,
        parents=[module.ap],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        )

    new_ap.set_defaults(func=module.main)


# arguments list
# a: angle unit
# d: trajectory output
# e: slice stop
# g: atom selections
# i: sort input by trail int
# l: selection
# o: topology output
# p: slice step
# r: reference frame
# s: slice start
# t: framelist
# plot: plot
# x: export data to table
# z: plane selection


def add_version_arg(parser):
    """
    Add version ``-v`` option to parser.
    
    Displays a message informing the current version.
    Also accessible via ``--version``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the version argument.
    """  # noqa: E501
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        # the _BANNER contains information on the version number
        version=(
            f'{_BANNER}\n'
            f'A record of the previous versions can be found at:\n'
            f'https://taurenmd.readthedocs.io/en/latest/changelog.html\n'
            )
        )


def add_angle_unit_arg(parser):
    """
    Add angle unit selectiona argument to parser.

    Is defined by ``-a`` and ``--aunit``.

    Wether angles are to be calculated in degrees or radians.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the topology positionl argument.
    """  # noqa: E501
    parser.add_argument(
        '-a',
        '--aunit',
        help='Angular unit, either degrees or radians.',
        choices=['degrees', 'radians'],
        default='degrees',
        )


def add_insort_arg(parser):
    """
    Sort input by trail int.

    Applies :py:func:`taurenmd.libs.libio.sort_numbered_input`.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the insort argument.
    """  # noqa: E501
    parser.add_argument(
        '-i',
        '--insort',
        help=(
            'Sorts input trajectories paths according to their tail numbers, '
            'if paths are formatted as follows: my_trajectory_#.dcd, '
            'where # is a number.'
            ),
        action='store_true',
        )


def add_topology_arg(parser):
    """
    Add topology positional argument to parser.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the topology positionl argument.
    """  # noqa: E501
    parser.add_argument(
        'topology',
        help='Path to the topology file.',
        type=str,
        )


def add_trajectories_arg(parser):
    """
    Add trajectory positional argument to parser.
    
    Accepts multiple trajectory files.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the trajectory positionl argument.
    """  # noqa: E501
    parser.add_argument(
        'trajectories',
        help=(
            'Path to the trajectory files. '
            'If multiple files are given, '
            'trajectories will be concatenated by input order.'
            ),
        nargs='+',
        )


def add_trajectory_arg(parser):
    """
    Add trajectory positional argument to parser.
    
    Accepts a single trajectory file.
    
    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the trajectory positionl argument.
    """  # noqa: E501
    parser.add_argument(
        'trajectory',
        help='Path to the trajectory file.',
        )


def add_slice_arg(parser):
    """
    Add start, stop and step slicing arguments.
    
    Slicing arguments are according to `Python Slice object <https://docs.python.org/3/library/functions.html#slice>`_

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to add the trajectory positionl argument.
    """  # noqa: E501
    parser.add_argument(
        '-s',
        '--start',
        help=(
            'The starting index for the frame slicing. '
            'Frames are 0-indexed, so the first frame is -s 0. '
            'The starting index is inclusive. '
            'Defaults to None, considers from the beginning.'
            ),
        default=None,
        type=int,
        )

    parser.add_argument(
        '-e',
        '--stop',
        help=(
            'The ending index for the frame slicing. '
            'Frames are 0-indexed, so the last frame of a 500 frame '
            'trajectory is index 499, but because '
            'the ending index is exclusive, -e 500 is required. '
            'Defaults to None, considers to the end.'
            ),
        default=None,
        type=int,
        )

    parser.add_argument(
        '-p',
        '--step',
        help=(
            'The periodicity step value for the frame slicing, '
            '-p 10 means every 10 frames. '
            'Defaults to None, considers every 1 frame.'
            ),
        default=None,
        type=int,
        )


def add_atom_selection_arg(parser):
    """
    Add selection optional argument.

    Selection argument is a string that defines the atom selection,
    this is defined by ``-l`` and ``--selection``, and defaults to ``all``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the selection argument.
    """  # noqa: E501
    parser.add_argument(
        '-l',
        '--selection',
        help=(
            'Atom selection for the output trajectory. '
            'Selection rules are as defined by the MD analysis '
            'library used by the client interface. '
            'For instructions read the main command-line client description. '
            'Defaults to \'all\'.'
            ),
        default='all',
        type=str,
        )


def add_atom_selections_arg(parser):
    """
    Add selections optional argument.

    Selections argument is a string that defines a list of atom selections,
    this is defined by ``-g`` and ``--selections``, and defaults to ``all``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the selections argument.
    """  # noqa: E501
    parser.add_argument(
        '-g',
        '--selections',
        help=(
            'List of atom selections to operate with. '
            'Selection rules are as defined by the MD analysis '
            'library used by the client interface. '
            'For instructions read the main command-line client description. '
            'Defaults to None, uses a single selection considering all '
            'atoms. '
            "Example: -g 'segid A' 'segid B' 'name CA'"
            ),
        default=None,
        nargs='+',
        )


def add_frame_list_arg(parser):
    """
    Add frame list argument.

    Registers a list of frame numbers, is defined by ``-t`` and ``--flist``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the flist argument.
    """  # noqa: E501
    parser.add_argument(
        '-t',
        '--flist',
        help=(
            'List of frames (time steps) to consider.'
            'If applicable, this can used instead of the start, stop '
            'and step slicing arguments.'
            ),
        default=False,
        nargs='+',
        type=int,
        )


def add_plane_selection_arg(parser):
    """
    Add plane selection argument.

    Plane selection is a selection of three regions separated by 'or'
    operator.
    
    Is defined by ``-z`` and ``--plane-selection``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the plane-selection argument.
    """  # noqa: E501
    parser.add_argument(
        '-z',
        '--plane-selection',
        help=(
            'Three selection strings representing three atom regions. '
            'The plane is defined by the three centres of geometry '
            'of the three selections. For example: '
            '-z \'segid A\' \'segid B\'  \'segid C\'.'
            ),
        required=True,
        nargs=3,
        )


def add_reference_frame_arg(parser):
    """
    Add a reference frame argument.

    Reference frame is the frame to compute the parameter against.

    Depending on the client logic the reference frame might have different
    meanings.

    Is defined by ``-r`` and ``--ref-frame``.

    Defaults to ``0``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the refence-frame argument.
    """  # noqa: E501
    parser.add_argument(
        '-r',
        '--ref-frame',
        help=(
            'The frame in the trajectory that serves as '
            'reference to compute against.'
            'Defaults to 0.'
            ),
        default=0,
        type=int,
        )


def add_plot_arg(parser):
    """
    Add plot parameters.

    Plot kwargs that will be passed to the plotting function.

    Defined by ``--plot``.

    If given, plot results. Additional arguments can be given to
    specify the plot parameters.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the plot argument.
    """  # noqa: E501
    parser.add_argument(
        '--plot',
        help=(
            'Plot results. '
            'Additional arguments can be given to configure the plot '
            'style. '
            'Example: --plot xlabel=frames ylabel=RMSD color=red.'
            'Accepted plot arguments are defined by the function used '
            'to plot the result. The main description of this client '
            'which plotting function is used. '
            'Defaults to ``None``, no plot is produced.'
            ),
        nargs='*',
        default=False,
        action=ParamsToDict,
        )


def add_top_output_arg(parser):
    """
    Add argument to export first frame as topology PDB file.

    Defined by ``-o`` and ``--top-output``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the topology output argument.
    """  # noqa: E501
    parser.add_argument(
        '-o',
        '--top-output',
        help=(
            'Export edited trajectory first frame as topololy file. '
            'You can specify the exact file name, otherwise, defaults to '
            'input trajectory path + \'_frame0.pdb\'. '
            'Also, if name starts with \'_\', it is used as file suffix, '
            'if name ends with \'_\', it is used as prefix, instead.'
            ),
        default=False,
        const='_frame0.pdb',
        nargs='?',
        )


def add_traj_output_arg(parser):
    """
    Add argument to export trajectory after client modifications.

    Defined by ``-d`` and ``--traj-output``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the trajectory output argument.
    """  # noqa: E501
    parser.add_argument(
        '-d',
        '--traj-output',
        help=(
            'Modified trajectory output file name. '
            'File type will be defined by file name extension. '
            'Defaults to traj_out.dcd.'
            ),
        default='traj_out.dcd',
        )


def add_data_export_arg(parser):
    """
    Add export argument.

    Export argument configures data export to a text file in table format.

    Is defined by ``-x`` and ``--export``.

    Parameters
    ----------
    parser : `argparse.ArgumentParser <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
        The argument parser to which add the export argument.
    """  # noqa: E501
    parser.add_argument(
        '-x',
        '--export',
        help=(
            'Export calculated values to a CSV file. '
            'Defaults to \'results.csv\', alternatively, '
            'you can give a specific file name.'
            ),
        default=False,
        const='results.csv',
        nargs='?',
        )



"""Test libmdt."""
import mdtraj as md
import pytest

from taurenmd.libs import libmdt

from . import toptest, toptest_cif, trajtest


def test_load_traj_pdb():
    """Test loading traj."""
    traj = libmdt.load_traj(toptest, trajtest)
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 10


def test_load_traj_cif():
    """Test loading traj."""
    traj = libmdt.load_traj(toptest_cif, trajtest)
    assert isinstance(traj, md.Trajectory)
    assert len(traj) == 10


def test_load_traj_cif_import_error():
    """Test loading traj."""
    libmdt.SIMTK = False
    with pytest.raises(SystemExit) as err:
        libmdt.load_traj(toptest_cif, trajtest)
    assert err.type == SystemExit
    assert err.value.code == 0


def test_simtk_import_error():
    """Test import error message."""
    with pytest.raises(SystemExit) as err:
        libmdt._log_simtkimport_error()
    
    assert err.type == SystemExit
    assert err.value.code == 0



"""Test things in init."""

import pytest

from taurenmd import _controlled_exit


def test_controlled_exit():
    """Test controlled exit function."""
    with pytest.raises(SystemExit) as exit:
        _controlled_exit()
    assert exit.type == SystemExit
    assert exit.value.code == 126



"""Handle input and output general operations."""
import re

from taurenmd import Path, log
from taurenmd.logger import S, T  # noqa: F401


def get_number(path):
    """
    Extract tail number from path.
    
    Examples
    --------

        >>> get_number('traj_1.dcd')
        >>> 1
        
        >>> get_number('traj_3.dcd')
        >>> 3
        
        >>> get_number('traj_1231.dcd')
        >>> 1231
        
        >>> get_number('traj_0011.dcd')
        >>> 11
        
        >>> get_number('traj_1_.dcd')
        >>> 1
        
        >>> get_number('traj_20200101_1.dcd')
        >>> 1

    Parameters
    ----------
    path : str or Path obj
        The path to evaluate.

    Returns
    -------
    int
        The tail integer of the path.
    """
    stem = Path(path).stem
    number = re.findall(r'\d+', stem)[-1]
    return int(number)


def sort_numbered_input(*inputs):
    """
    Sort input paths to tail number.

    If possible, sort criteria is provided by :py:func:`get_number`.
    If paths do not have a numbered tag, sort paths alphabetically.

    Parameters
    ----------
    *inputs : str of Paths
        Paths to files.

    Returns
    -------
    list
        The sorted pathlist.
    """
    try:
        return sorted(inputs, key=get_number)
    except TypeError as err:
        log.exception(err)
        emsg = "Mind the packing *argument, input should be strings or Paths"
        raise TypeError(emsg)
    except IndexError as err:
        log.exception(err)
        log.info(S('sorting done by string type'))
        return sorted(inputs)

   
def report_input(topology, trajectory):
    """Report on topology and trajectory file paths used as input."""
    log.info(S('loading trajectory: {}', trajectory))
    log.info(S('with topology: {}', topology))
    return


def add_prefix_to_path(ipath, prefix):
    """
    Add prefix to file path.
    
    Example
    -------

        >>> mk_frame_path('traj_output.xtc', prefix='my_prefix')
        >>> my_prefixtraj_output.xtc

    Mind the ``_`` is not placed automatically.

        >>> mk_frame_path('traj_output.xtc', prefix='my_prefix_')
        >>> my_prefix_traj_output.xtc

    Parameters
    ----------
    ipath : str or Path
        The file path to alter.

    prefix : str
        The complete prefix for the file name.

    Returns
    -------
    :py:func:`taurenmd.core.Path`
        The new file path.
    """
    ipath_ = _get_ipath(ipath)
    return Path(
        ipath_.myparents(),
        '{}{}'.format(prefix, ipath_.name),
        )


def add_suffix_to_path(ipath, suffix):
    """
    Add suffix to file path.
    
    If suffix has extention, updates the path extension, otherwise
    keeps the original extension.
    
    Examples
    --------

        >>> mk_frame_path('traj_output.xtc', suffix='my_suffix')
        >>> traj_outputmy_suffix.xtc

    Mind the underscore is not placed automatically:

        >>> mk_frame_path('traj_output.xtc', suffix='_my_suffix')
        >>> traj_output_my_suffix.xtc
   
    Updating extensions:

        >>> mk_frame_path('traj_output.xtc', suffix='_my_suffix.pdb')
        >>> traj_output_my_suffix.pdb

    Parameters
    ----------
    ipath : str or Path
        The file path to alter.

    suffix : str
        The complete suffix for the file name, extension should be
        included in the suffix, extension of the ``ipath`` is
        ignored.

    Returns
    -------
    :py:func:`taurenmd.core.Path`
        The new file path.
    """
    ipath_ = _get_ipath(ipath)
    result = Path(
        ipath_.myparents(),
        '{}{}'.format(ipath_.stem, suffix),
        )
    
    extension = Path(suffix).suffix or ipath_.suffix
    return result.with_suffix(extension)


def mk_frame_path(ipath, frame=0, ext='.pdb', leading=0, suffix=None):
    """
    Create the path name for a frame.

    Given an input_path ``ipath`` (usually the name of the trajectory),
    create the corresponding frame file name.

    Example
    -------

        >>> mk_frame_path('traj_output.xtc')
        >>> traj_output_frame0.pdb
       
        >>> mk_frame_path('traj_output.xtc', frame=4, leading=4)
        >>> traj_output_frame0004.pdb
   
    Parameters
    ----------
    ipath : str or Path
        The file path. Normally, trajectory file path.

    frame : int, optional
        The frame to label the new path.
        Defaults to ``0``.

    ext : str, optional
        The returned file desired extension.
        Defaults to ``.pdb``.

    leading : int
        The leading zeros to left append to the frame number.
        Defaults to ``0``.
    
    suffix : str
        Complete specifications of the desired suffix.
        If ``suffix`` is given, ``frame`` and ``ext`` and ``leading``
        are ignored and :py:func:`add_suffix_to_path` is used directly.

    Returns
    -------
    :py:func:`taurenmd.core.Path`
        The new file path.
    """
    if suffix:
        return add_suffix_to_path(ipath, suffix)
    else:
        ipath_ = _get_ipath(ipath)
        return Path(
            ipath_.myparents(),
            '{}_frame{}'.format(ipath_.stem, str(frame).zfill(leading)),
            ).with_suffix('.{}'.format(ext.lstrip('.')))


def _get_ipath(ipath):
    try:
        return Path(ipath)
    except TypeError as err:
        log.exception(err)
        return Path()


def parse_top_output(top_output, traj_output=None):
    """
    Parse different output definitions for topology output file name.
   
    If ``top_output`` startswith ``_`` uses :py:func:add_suffix_to_path.
    If ``top_output`` endswith ``_`` uses :py:func:add_prefix_to_path.
    Else: Return Path object of ``top_output``.


    Parameters
    ----------
    top_output : str or Path
        The string to evaluate.

    traj_output : str or Path
        The trajectory output file name. Return value depends on
        this parameters.

    Returns
    -------
    :py:func:`taurenmd.core.Path`
        The new topology file path.
    """
    if str(top_output).startswith('_'):
        return add_suffix_to_path(traj_output, suffix=top_output)

    elif str(top_output).endswith('_'):
        return add_prefix_to_path(traj_output, prefix=top_output)

    else:
        return Path(top_output)


def export_data_to_file(
        xdata,
        *ydata,
        fname='results.csv',
        header='',
        fmt='{:.3f}',
        delimiter=',',
        ):
    """
    Save data to file.

    Following the format:

        >>> # header
        >>> x1,ya1,yb1,yc1
        >>> x2,ya2,yb2,yc2
        >>> x3,ya3,yb3,yc3

    Where ``x`` are the elements in ``xdata``, and ``y*`` are the
    elements in the different ``ydata`` series.

    Parameters
    ----------
    xdata : list-like
        Contains the *x axis* data.

    ydata : list of list-like
        Contains the *y axis* data series.

    fname : str
        The output file path name.

    header : str
        The commented header of the file. Comment character, like ``#``
        is not placed, it should be already provided if desired.

    fmt : str
        The float format. Defaults to ``{:.3f}``.

    delimiter : str
        The string delimiter between columns. Defaults to ``,``.
    """
    with open(fname, 'w') as fh:
        fh.write(header)
        for label, ydataseries in zip(
                xdata,
                zip(*ydata),
                ):

            fh.write('{}{}{}\n'.format(
                label,
                delimiter,
                delimiter.join(fmt.format(float(i)) for i in ydataseries),
                ))
    return


def frame_list(len_traj, start=None, stop=None, step=None, flist=None,):
    """
    Create frame integer list from a length and slice parameters.
    
    Examples
    --------

        >>> frame_list(10)
        >>> list(range(10)) # returns

        >>> frame_list(10, start=2, stop=5)
        >>> list(range(2, 5)) # returns

        >>> frame_list(10, 2, 50)
        >>> list(range(2, 10)) # returns

        >>> frame_list(10, None, None, 2)
        >>> list(range(0, 10, 2)) # returns

        >>> frame_list(10, flist='1,2,45,65')
        >>> [1, 2, 45, 65]

        >>> frame_list(10, flist=[1, 2, 45, 65])
        >>> [1, 2, 45, 65]

        >>> frame_list(10, flist=['1', '2', '45', '65'])
        >>> [1, 2, 45, 65]
        
        >>> frame_list(None, flist=['1', '2', '45', '65'])
        >>> [1, 2, 45, 65]

    Parameters
    ----------
    len_traj : int
        The length to evaluate. Normally this is the length of the
        trajectory.

    start : int or None, optional
        The start index for the frame list after length evaluation.
        Defaults to ``None``.
    
    stop : int or None, optional
        The stop index for the frame list after length evaluation.
        Defaults to ``None``.

    step: int or None, optional
        the step index for the frame list after length evaluation.
        Defaults to ``None``.

    flist : list-like, or comma-separated string, optional
        The list of specific frames.
        Defaults to ``None``.

    Returns
    -------
    list
        The resulting frame list.
    """
    if any((start, stop, step)):
        return list(range(len_traj)[slice(start, stop, step)])

    elif flist:
        try:
            return [int(i) for i in flist.split(',')]
        except AttributeError:
            try:
                return [int(i) for i in flist]
            except TypeError:
                raise ValueError(
                    'Cannot generate list of frames from {}'.format(flist)
                    ) from None
    else:
        return list(range(len_traj))
   

def frame_slice(start=None, stop=None, step=None,):
    """
    Create a slice object from parameters.

    Logs the created slice object.

    Returns
    -------
    Slice Object
        The slice object created from slice(start, stop, step)
    """
    log.info(S('slicing: {}:{}:{}', start, stop, step))
    return slice(start, stop, step)


def evaluate_to_slice(*, value=None, start=None, stop=None, step=None):
    """
    Evaluate to slice.
    
    If any of ``start``, ``stop`` or ``step`` is given returns
    ``slice(start, stop, step)``. Otherwise tries to evaluate ``value``
    to its representative slice form.

    Examples
    --------

        >>> evaluate_to_slice(value='1,100,2')
        >>> slice(1, 100, 2)
        
        >>> evaluate_to_slice(start=10)
        >>> slice(10, None, None)

        >>> evaluate_to_slice(value=(0, 50, 3))
        >>> slice(0, 50, 3)

        >>> evaluate_to_slice(value=(None, 100, None))
        >>> slice(None, 100, None)

        >>> #ATTENTION
        >>> evaluate_to_slice(value='10')
        >>> slice(10, None, None)
        >>> # this is different from slice(10)
        >>> #though
        >>> evaluate_to_slice(value=10)
        >>> slice(None, 10, None)

    Parameters
    ----------
    value : list, tuple, str, None or int
        A human readable value that can be parsed to a slice object
        intuitively.
        Defaults to ``None``.

    start : None or int
        The starting index of the slice (inclusive).
        Defaults to ``None``.

    stop : None or int
        The final index of the slice (exclusive).
        Defaults to ``None``.

    step : None or int
        Slice periodicity.
        Defaults to ``None``.
    
    Returns
    -------
    slice
        Python `slice object <https://docs.python.org/3/library/functions.html#slice>`_.

    Raises
    ------
    ValueError
        If slice can not be computed.
    """  # noqa: E501
    def convert(a):
        try:
            return int(a)
        except (ValueError, TypeError):
            return None
    
    def eval_types(list_):
        types = (int, str, type(None))
        return all(isinstance(i, types) for i in list_)

    if any((start, stop, step)) and eval_types([start, stop, step]):
        return slice(convert(start), convert(stop), convert(step))

    elif isinstance(value, (list, tuple)) \
            and len(value) == 3 \
            and eval_types(value):
        return slice(convert(value[0]), convert(value[1]), convert(value[2]))

    elif value is None:
        return slice(None, None, None)

    elif isinstance(value, slice):
        return value

    elif isinstance(value, str):
        if value.find(':') > -1:
            values = value.split(':')
        elif value.find(',') > -1:
            values = value.split(',')
        else:
            values = value.split()
        
        indexes = [start, stop, step]
        for i, v in enumerate(values):
            indexes[i] = convert(v)
            
        return slice(*indexes)
    
    elif isinstance(value, int):
        return slice(value)

    else:
        raise ValueError('slice object could not be generated')



"""Test taurenmd client interfaces."""
import argparse
import copy
import functools
import inspect
import sys

import pytest

# import bellow by alphabetical order the cli interfaces you are
# implementing
from taurenmd import __main__ as main_module
from taurenmd import (  # noqa: E133
    cli,
    cli_distances,
    cli_fext,
    cli_imagemol,
    cli_nosol,
    cli_pangle,
    cli_report,
    cli_rmsd,
    cli_rmsf,
    cli_rotations,
    cli_trajedit,
)
from taurenmd.libs import libcli as lc

from . import Path, toptest, trajtest


subclients = [
    # add your new client to this list.
    cli_distances,
    cli_fext,
    cli_imagemol,
    cli_nosol,
    cli_pangle,
    cli_report,
    cli_rmsd,
    cli_rmsf,
    cli_rotations,
    cli_trajedit,
    ]


@pytest.fixture(params=subclients)
def client(request):
    """Loop all individual clients."""
    return request.param


@pytest.fixture(params=[cli] + subclients)
def all_clients(request):
    """Consider individual clientes plus main cli."""
    return request.param


def test_main_module():
    """Test main module entry point."""
    assert hasattr(main_module, 'maincli')


def test_main_module_if():
    """Test cli main entry point."""
    lines = inspect.getsource(main_module).split('\n')
    assert lines[-3] == "if __name__ == '__main__':"
    assert lines[-2] == "    maincli()"
    assert lines[-1] == ''


def test_cli__version(all_clients):
    """Test clients have version parameter."""
    with pytest.raises(SystemExit) as err:
        all_clients.ap.parse_args(['-v'])
    assert err.value.code == 0


def test_cli_script_1():
    """Test cli main entry point."""
    assert hasattr(cli, 'maincli')


def test_cli_script_2():
    """Test cli has load_args function."""
    assert hasattr(cli, 'load_args')


def test_cli_script_3():
    """Test cli has _ap function."""
    assert hasattr(cli, '_ap')


def test_cli__ap_returns():
    """Test cli _ap returns ap."""
    assert isinstance(cli._ap(), argparse.ArgumentParser)


def test_cli_script_4():
    """Test cli main if __name__ part."""
    lines = inspect.getsource(cli).split('\n')
    assert lines[-3] == "if __name__ == '__main__':"
    assert lines[-2] == "    maincli()"
    assert lines[-1] == ''


def test_cli_load_args():
    """Test cli load_args functions."""
    backup = copy.deepcopy(sys.argv)
    sys.argv = ['taurenmd', 'report', toptest.str(), trajtest.str()]
    result = cli.load_args()
    assert isinstance(result, argparse.Namespace)
    sys.argv = copy.deepcopy(backup)


def test_cli_maincli_execution():
    """Test cli maincli execution."""
    backup = copy.deepcopy(sys.argv)
    sys.argv = ['taurenmd', 'report', toptest.str(), trajtest.str()]
    cli.maincli()
    sys.argv = copy.deepcopy(backup)


def test_cli_maincli_SystemExit():
    """Test cli.py maincli() SystemExit."""
    backup = copy.deepcopy(sys.argv)
    sys.argv = ['taurenmd']
    with pytest.raises(SystemExit):
        cli.maincli()
    sys.argv = copy.deepcopy(backup)


def test_clients_have_maincli(client):
    """Test all modules have maincli."""
    assert hasattr(client, 'maincli')


def test_clientes_have__ap(client):
    """Test all modules have _ap."""
    assert hasattr(client, '_ap')


def test_clients_have__ap_returns(client):
    """Test all modules _ap() return ap."""
    assert isinstance(client._ap(), argparse.ArgumentParser)


def test_clients_have_help(client):
    """Test all modules have help."""
    assert hasattr(client, '_help')


def test_clients_have_main(client):
    """Test all modules have main."""
    assert hasattr(client, 'main')


def test_clients_if_main_code(client):
    """Test if __name__ == '__main__' code lines."""
    lines = inspect.getsource(client).split('\n')
    assert lines[-3] == "if __name__ == '__main__':"
    assert lines[-2] == "    maincli()"
    assert lines[-1] == ''


def test_clients_have_ap(client):
    """Test all modules have ap."""
    assert hasattr(client, 'ap')


def test_clients_main_partial_type(client):
    """Test maincli partial type."""
    assert isinstance(client.maincli, functools.partial)


def test_clients_maincli_partial_func(client):
    """Test maincli partial args."""
    assert client.maincli.func == (lc.maincli)


def test_clients_maincli_partial_args(client):
    """Test maincli partial args."""
    assert client.maincli.args == (client.ap, client.main)


@pytest.mark.parametrize(
    'module,name',
    [
        # add here the name of your client, this ensures that changes
        # in the command interface get noticed by the tests.
        # (cli_NAME, 'NAME'),
        (cli_distances, 'dist'),
        (cli_fext, 'fext'),
        (cli_imagemol, 'imagemol'),
        (cli_nosol, 'nosol'),
        (cli_pangle, 'pangle'),
        (cli_report, 'report'),
        (cli_rmsd, 'rmsd'),
        (cli_rmsf, 'rmsf'),
        (cli_rotations, 'rotations'),
        (cli_trajedit, 'trajedit'),
        ],
    )
def test_cli__name(module, name):
    """Test name messages."""
    assert module._name == name


# write a test function that calls the main function for your client.
# and add any additional modification or output confirmation that you
# find appropriate so that test coverage is 100% and you asure the
# correct output is generate. Read through the other client for examples.
# there is a final test to add at the end of the file, do not miss it.

# def test_cli_NAME_1():
#     """Test cli NAME."""
#     cli_NAME.main(
#         #params...
#         )


def test_cli_distances_1():
    """Test cli distances."""
    cli_distances.main(
        toptest,
        [trajtest],
        sel1='resnum 10',
        sel2='resnum 20',
        start=5,
        end=None,
        step=None,
        export='dist.csv',
        plot=True,
        plotvars={'filename': 'dist.pdf'},
        )
    Path('dist.csv').unlink()
    Path('dist.pdf').unlink()


def test_cli_distances_2():
    """Test cli distances."""
    cli_distances.main(
        toptest,
        [trajtest],
        sel1='resnum 10',
        sel2='resnum 20',
        start=None,
        end=5,
        step=None,
        export=False,
        plot=False,
        )
    assert not Path('dist.csv').exists()
    assert not Path('dist.pdf').exists()


def test_cli_fext_1():
    """Test client frame extract."""
    cli_fext.main(
        toptest,
        [trajtest],
        start=4,
        stop=6,
        step=None,
        prefix='timestep_',
        ext='pdb',
        selection='name CA',
        )
    p1 = Path('timestep_04.pdb')
    p2 = Path('timestep_05.pdb')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_imagemol():
    """Test imagemol."""
    cli_imagemol.main(
        toptest,
        trajtest,
        traj_output='testimage.dcd',
        top_output='testtop.pdb',
        protocol=1,
        )
    p1 = Path('testimage.dcd')
    p2 = Path('testtop.pdb')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_imagemol_2():
    """Test imagemol."""
    cli_imagemol.main(
        toptest,
        trajtest,
        traj_output='testimage.dcd',
        top_output=False,
        protocol=2,
        )
    p1 = Path('testimage.dcd')
    assert p1.exists()
    p1.unlink()


def test_cli_nosol_1():
    """Test cli nosol."""
    cli_nosol.main(
        toptest,
        trajtest,
        selection='name CA',
        maintain=None,
        top_output='_f0.pdb',
        traj_output='nosol.dcd',
        )
    p1 = Path('nosol.dcd')
    p2 = Path('nosol_f0.pdb')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_nosol_2():
    """Test cli nosol."""
    cli_nosol.main(
        toptest,
        trajtest,
        selection=None,
        maintain=['CL'],
        top_output='_f1.pdb',
        traj_output='nosol.dcd',
        )
    p1 = Path('nosol.dcd')
    p2 = Path('nosol_f1.pdb')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_nosol_3():
    """Test cli nosol."""
    cli_nosol.main(
        toptest,
        trajtest,
        selection=None,
        maintain=['CL', 'NA'],
        top_output=None,
        traj_output='nosol.dcd',
        )
    p1 = Path('nosol.dcd')
    p2 = Path('nosol_f1.pdb')
    assert p1.exists()
    assert not p2.exists()
    p1.unlink()


def test_cli_pangle_1():
    """Test client plane angle."""
    cli_pangle.main(
        toptest,
        [trajtest],
        ['resnum 10', 'resnum 20', 'resnum 30'],
        aunit='degrees',
        export='angles.csv',
        plot=True,
        plotvars={'filename': 'angles.pdf'},
        )
    p1 = Path('angles.csv')
    p2 = Path('angles.pdf')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_pangle_2():
    """Test client plane angle."""
    cli_pangle.main(
        toptest,
        [trajtest],
        ['resnum 10', 'resnum 20', 'resnum 30'],
        aunit='radians',
        export=False,
        plot=False,
        )


def test_cli_report_1():
    """Test cli report."""
    cli_report.main(toptest, [trajtest])


def test_cli_rmsd_1():
    """Test cli rmsd."""
    cli_rmsd.main(
        toptest,
        [trajtest],
        selections=['name CA'],
        )


def test_cli_rmsd_2():
    """Test cli rmsd."""
    cli_rmsd.main(
        toptest,
        [trajtest],
        selections=None,
        export='rmsds.csv',
        plot=True,
        plotvars={'filename': 'rmsds.pdf'},
        )
    p1 = Path('rmsds.csv')
    p2 = Path('rmsds.pdf')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_rmsf_1():
    """Test cli rmsf."""
    cli_rmsf.main(
        toptest,
        [trajtest],
        selections=['name CA'],
        )


def test_cli_rmsf_2():
    """Test cli rmsd."""
    cli_rmsf.main(
        toptest,
        [trajtest],
        selections=None,
        export='rmsfs.csv',
        plot=True,
        plotvars={'filename': 'rmsfs.pdf'},
        )
    p1 = Path('rmsfs.csv')
    p2 = Path('rmsfs.pdf')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_rmsf_error1():
    """Test cli rmsd."""
    with pytest.raises(TypeError):
        cli_rmsf.main(
            toptest,
            [trajtest],
            selections='name CA',
            )


def test_cli_rmsf_error2():
    """Test cli rmsd."""
    with pytest.raises(TypeError):
        cli_rmsf.main(
            toptest,
            [trajtest],
            selections=[],
            )


def test_cli_rotations_1():
    """Test cli rotations."""
    cli_rotations.main(
        toptest,
        [trajtest],
        plane_selection=['resnum 10', 'resnum 20', 'resnum 30'],
        aunit='radians',
        export='rots.csv',
        )

    p1 = Path('roll_angles_rots.csv')
    p2 = Path('pitch_angles_rots.csv')
    p3 = Path('yaw_angles_rots.csv')
    assert p1.exists()
    assert p2.exists()
    assert p3.exists()
    p1.unlink()
    p2.unlink()
    p3.unlink()


def test_cli_rotations_2():
    """Test cli rotations."""
    cli_rotations.main(
        toptest,
        [trajtest],
        plane_selection=['resnum 10', 'resnum 20', 'resnum 30'],
        )


def test_cli_trajedit_1():
    """Test trajedit 1."""
    cli_trajedit.main(
        toptest,
        [trajtest],
        insort=True,
        selection='name CA',
        traj_output='traj.dcd',
        top_output='topooo.pdb',
        unwrap=True,
        align='name CA',
        )
    p1 = Path('traj.dcd')
    p2 = Path('topooo.pdb')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()


def test_cli_trajedit_SystemExit_align():
    """Test trajedit 1."""
    with pytest.raises(SystemExit) as exit:
        cli_trajedit.main(
            toptest,
            [trajtest],
            align='name ZZZ',
            )
    assert exit.type == SystemExit
    assert exit.value.code == 126


def test_cli_trajedit_2():
    """Test trajedit 2."""
    cli_trajedit.main(
        toptest,
        [trajtest],
        )
    p1 = Path('traj_out.dcd')
    assert p1.exists()
    p1.unlink()


def test_cli_trajedit_3():
    """Test trajedit 3."""
    cli_trajedit.main(
        toptest,
        [trajtest],
        top_output='topooo.pdb',
        )
    p2 = Path('topooo.pdb')
    p1 = Path('traj_out.dcd')
    assert p1.exists()
    assert p2.exists()
    p1.unlink()
    p2.unlink()

# test argparse


@pytest.mark.parametrize(
    'module,cmd',
    [
        # add command examples to this list
        (cli, 'fext top.pdb traj1.xtc traj2.xtc'),
        (cli, 'dist top.pdb traj1.xtc traj2.xtc -l1 selA -l2 selB -s 1 -e 2 -p 3 -x --plot'),  # noqa: E501
        (cli, 'imagemol top.pdb traj.xtc -d tout.xtc -o'),
        (cli, 'imagemol top.pdb traj.xtc -d tout.xtc -o out.pdb'),
        (cli, 'pangle top.pdb traj_1.xtc traj_2.xtc -x --plot title=title -s 1 -e 2 -p 3 -z selA selB selC -a radians'),  # noqa: E501
        (cli, 'trajedit top.pdb traj1.xtc traj2.xtc -i -l segA -s 1 -e 10 -p 2 -d tout.xtc -o'),  # noqa: E501
        (cli, 'nosol top.pdb traj.xtc -d tout.xtc -o out.pdb -m NA'),
        (cli, 'rmsd top.pdb traj1.xtc traj2.xtc -g segA segB -s 1 -e 10 -p 2 -r 10 -x data.csv --plot'),  # noqa: E501
        (cli, 'rmsd top.pdb traj1.xtc traj2.xtc -g segA segB -s 1 -e 10 -p 2 -r 10 -x data.csv --plot title=1'),  # noqa: E501
        (cli, 'rmsf top.pdb traj1.xtc traj2.xtc -g segA segB -s 1 -e 10 -p 2 -x data.csv --plot title=1'),  # noqa: E501
        (cli, 'rotations top.pdb traj1.xtc traj2.xtc -z segA segB segC -s 1 -e 10 -p 2 -a radians -x data.csv'),  # noqa: E501
        (cli, 'report top.pdb traj1.xtc traj2.xtc'),
        ],
    )
def test_ap_interface(module, cmd):
    """Test argparse interface of command lines."""
    module.ap.parse_args(cmd.split())



"""
# Calculate RMSFS of a selection along the trajectory slice.

## Algorithm

Calculates the RMSF values along a trajectory slice for different
selections. If multiple selections are given creates a series data
for that selection.

RMSF is calculated using [libcalc.mda_rmsf](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.mda_rmsf).

If multiple selections are given, separate calculations are performed
in sequence. Result files (data tables and plots) are exported
separately for each selection. Selections can't be overlayed easily
in a single plot because they do not share the same labels.

## Examples

Calculate RMSF of the whole system:

    taurenmd rmsf top.pdb traj.dcd -e rmsf.csv

Calculates RMSFs for different selections:

    taurenmd rmsf top.pdb traj.dcd -g 'segid A' 'segid B' -e

``-x`` exports the data to a CSV file. You can also plot the data with
the ``-v`` option:

    [...] -x rmsf.csv -v title=my-plot-title xlabel=frames ylabel=RMSFs ...

where ``[...]`` is the previous command example.

you can also use ``tmdrmsf`` instead of ``taurenmd rmsf``.

## References

"""  # noqa: E501
import argparse
import functools
from datetime import datetime

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log  # noqa: F401
from taurenmd.libs import libcalc, libcli, libio, libmda, libplot
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_plottemplates_labeldots}'
    )

_help = 'Calculate RMSFS for a selection and trajectory slice.'
_name = 'rmsf'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_atom_selections_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        start=None,
        stop=None,
        step=None,
        selections=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs
        ):
    """Execute client main logic."""
    log.info(T('starting RMSFs calculation'))
    
    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    u = libmda.load_universe(topology, *trajectories)
   
    frame_slice = libio.frame_slice(
        start=start,
        stop=stop,
        step=step,
        )
  
    if selections is None:
        selections = ['all']
    
    if not isinstance(selections, list) or len(selections) == 0:
        raise TypeError('selections must be LIST with at least one element')
    
    log.info(T('calculating'))
    for sel in selections:
        labels = []
        rmsfs = []
        log.info(S('for selection: {}', sel))
        atom_group = u.select_atoms(sel)
        labels = libmda.draw_atom_label_from_atom_group(atom_group)
       
        rmsfs = libcalc.mda_rmsf(
            atom_group,
            frame_slice=frame_slice,
            )

        if export:
            libio.export_data_to_file(
                labels,
                rmsfs,
                fname=export,
                header=(
                    "# Date: {}\n"
                    "# Topology: {}\n"
                    "# Trajectories {}\n"
                    "# Atom,RMSF\n"
                    ).format(
                        datetime.now().strftime("%d/%m/%Y, %H:%M:%S"),
                        Path(topology).resolve(),
                        ', '.join(f.resolve().str() for f in trajectories),
                        ),
                )
    
        if plot:
            plotvars = plotvars or dict()
            plotvars.setdefault('series_labels', selections)

            log.info(T('plot params:'))
            for k, v in plotvars.items():
                log.info(S('{} = {!r}', k, v))
            
            libplot.label_dots(
                labels,
                rmsfs,
                **plotvars,
                )

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""Test libcli."""
import argparse

import pytest

from taurenmd import Path, references
from taurenmd.libs import libcli as lc
from taurenmd.logger import CMDFILE


def test_load_args():
    """Test load_args func."""
    parser = argparse.ArgumentParser()
    # this has to --cov because otherwise argparse catches the
    # pytest commands
    parser.add_argument('--cov', nargs=argparse.REMAINDER)
    result = lc.load_args(parser)
    assert isinstance(result, argparse.Namespace)


def test_maincli():
    """Test maincli."""
    def myfunc(cov, *args, **kwargs):
        return cov
    parser = argparse.ArgumentParser()
    parser.add_argument('--cov', nargs=argparse.REMAINDER)
    result = lc.maincli(parser, myfunc)
    # since we have it, lets play it with and close the circle
    assert result == (
        '--cov-report=term-missing '
        '--cov-append --cov-config=.coveragerc -vv tests'
        ).split()


def test_save_refs():
    """Test save references."""
    references.add("zello world")
    lc.save_references()
    p1 = Path(CMDFILE)
    s = p1.open().readlines()
    assert s[-1].split(':')[1][1:] == 'zello world'
    p1.unlink()


def test_CustomParser():
    """Test Custom Parser."""
    assert issubclass(lc.CustomParser, argparse.ArgumentParser)
    assert hasattr(lc.CustomParser, 'error')


def test_CustomParser_error():
    """Test CP error."""
    with pytest.raises(SystemExit) as error:
        lc.CustomParser().error('my error')
    assert error.type == SystemExit
    assert error.value.code == 2


@pytest.mark.parametrize(
    'key,value,expected',
    [
        ('tilte=', 'my title', 'my title'),
        ('colors=', 'k,b,y,m', ('k', 'b', 'y', 'm')),
        ('xrange=', '1.5,100', (1.5, 100)),
        ('xmin=', '1.5', 1.5),
        ('xmin=', '15', 15),
        ('var1', '', True),
        ],
    )
def test_ParamsToDick(key, value, expected):
    """Test params to dict parsing."""
    p = lc.ParamsToDict(dest='plot', option_strings='--plot')
    namespace = argparse.Namespace
    parser = argparse.ArgumentParser()
    p(
        parser,
        namespace,
        [f'{key}{value}'],
        )
    v = vars(namespace)

    assert 'plotvars' in v
    assert v['plotvars'][key.rstrip('=')] == expected


def test_save_command():
    """Tests only the interface."""
    lc.save_command('testcommandsave', 1, 2, 3, 4)
    Path('testcommandsave').unlink()


def test_add_subparser():
    """Test adds subparser."""

    def mainfunc():
        return 'this is main func'

    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers()
    parserm = argparse.ArgumentParser()
    parserm.add_argument('foo')
    parserm.add_argument('--bar', nargs=1, default=False)
    pseudomodule = argparse.Namespace(
        _name='mycmd',
        __doc__='Module documentation',
        _help='short help',
        ap=parserm,
        main=mainfunc,
        )

    lc.add_subparser(subparser, pseudomodule)
    v = vars(parser.parse_args('mycmd this_foo --bar b'.split()))
    assert v['foo'] == 'this_foo'
    assert v['bar'] == ['b']


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-a degrees', 'degrees'),
        ('--aunit degrees', 'degrees'),
        ('-a radians', 'radians'),
        ('--aunit radians', 'radians'),
        ('', 'degrees'),
        ]
    )
def test_angle_unit(cmd, expected):
    """Test angle unit argument."""
    parser = argparse.ArgumentParser()
    lc.add_angle_unit_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['aunit'] == expected


def test_angle_unit_error():
    """Test sysexit error."""
    parser = argparse.ArgumentParser()
    lc.add_angle_unit_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args('-a'.split())


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-i', True),
        ('--insort', True),
        ('', False),
        ]
    )
def test_insort(cmd, expected):
    """Test angle unit argument."""
    parser = argparse.ArgumentParser()
    lc.add_insort_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['insort'] == expected


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('topology.pdb', 'topology.pdb'),
        ]
    )
def test_topology_input(cmd, expected):
    """Test topology argument."""
    parser = argparse.ArgumentParser()
    lc.add_topology_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['topology'] == expected


def test_topology_input_error():
    """Test topology input error."""
    parser = argparse.ArgumentParser()
    lc.add_topology_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args([])


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('traj1.dcd', ['traj1.dcd']),
        ('traj1.dcd traj2.dcd', ['traj1.dcd', 'traj2.dcd']),
        ]
    )
def test_trajectories(cmd, expected):
    """Test trajectories argument."""
    parser = argparse.ArgumentParser()
    lc.add_trajectories_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['trajectories'] == expected


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('traj1.dcd', 'traj1.dcd'),
        ]
    )
def test_trajectory(cmd, expected):
    """Test trajectory argument."""
    parser = argparse.ArgumentParser()
    lc.add_trajectory_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['trajectory'] == expected


def test_trajectory_error():
    """Test error in trajectory input."""
    parser = argparse.ArgumentParser()
    lc.add_trajectory_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args('traj1.dcd traj2.dcd'.split())


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-s 1 -e 100 -p 4', (1, 100, 4)),
        ('-e 100 -p 4', (None, 100, 4)),
        ('-s 100 -p 4', (100, None, 4)),
        ('-s 100 -e 4', (100, 4, None)),
        ]
    )
def test_slice_arg(cmd, expected):
    """Test slice arguments."""
    parser = argparse.ArgumentParser()
    lc.add_slice_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['start'] == expected[0]
    assert v['stop'] == expected[1]
    assert v['step'] == expected[2]


@pytest.mark.parametrize(
    'cmd,expected',
    [
        (['-l', "name CA"], 'name CA'),
        (['--selection', "name CA and segid A"], 'name CA and segid A'),
        ([], 'all'),
        ],
    )
def test_atom_selection_args(cmd, expected):
    """Test atom selection arg."""
    parser = argparse.ArgumentParser()
    lc.add_atom_selection_arg(parser)
    v = vars(parser.parse_args(cmd))
    assert v['selection'] == expected


@pytest.mark.parametrize(
    'cmd',
    [
        ('-l selA selB'),
        ('--selection selB selC selD'),
        ],
    )
def test_atom_selection_error(cmd):
    """Test atom selection error."""
    parser = argparse.ArgumentParser()
    lc.add_atom_selection_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd.split())


@pytest.mark.parametrize(
    'cmd,expected',
    [
        (['-g', "name CA"], ['name CA']),
        (['--selections', "name CA", "segid A"], ['name CA', 'segid A']),
        ([], None),
        ],
    )
def test_atom_selections_args(cmd, expected):
    """Test atom selection arg."""
    parser = argparse.ArgumentParser()
    lc.add_atom_selections_arg(parser)
    v = vars(parser.parse_args(cmd))
    assert v['selections'] == expected


@pytest.mark.parametrize(
    'cmd',
    [
        ('-g'),
        ('--selections'),
        ],
    )
def test_atom_selections_error(cmd):
    """Test atom selection error."""
    parser = argparse.ArgumentParser()
    lc.add_atom_selections_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd.split())


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-t 1', [1]),
        ('--flist 1 2', [1, 2]),
        ],
    )
def test_framelist_arg(cmd, expected):
    """Test frame list."""
    parser = argparse.ArgumentParser()
    lc.add_frame_list_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['flist'] == expected


def test_framelist_error():
    """Test framelist error."""
    parser = argparse.ArgumentParser()
    lc.add_frame_list_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args(['-t'])


@pytest.mark.parametrize(
    'cmd,expected',
    [
        (
            ['-z', 'segid A', 'segid B', 'segid C'],
            ['segid A', 'segid B', 'segid C']
            ),
        (
            ['--plane-selection', 'segid A', 'segid B', 'segid C'],
            ['segid A', 'segid B', 'segid C']
            ),
        ],
    )
def test_plane_selection(cmd, expected):
    """Test plane selection."""
    parser = argparse.ArgumentParser()
    lc.add_plane_selection_arg(parser)
    v = vars(parser.parse_args(cmd))
    assert v['plane_selection'] == expected


@pytest.mark.parametrize(
    'cmd',
    [
        (['-z']),
        (['-z', 'A']),
        (['-z', 'A', 'B']),
        (['-z', 'A', 'B', 'D', 'E']),
        ],
    )
def test_plane_selection_error(cmd):
    """Test plane selection error."""
    parser = argparse.ArgumentParser()
    lc.add_plane_selection_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd)


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-r 1', 1),
        ('--ref-frame 2', 2),
        ('', 0),
        ],
    )
def test_reference_frame(cmd, expected):
    """Test reference frame."""
    parser = argparse.ArgumentParser()
    lc.add_reference_frame_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['ref_frame'] == expected


@pytest.mark.parametrize(
    'cmd',
    [
        (['-r', 'A', 'B', 'D', 'E']),
        ],
    )
def test_reference_frame_error(cmd):
    """Test plane selection error."""
    parser = argparse.ArgumentParser()
    lc.add_reference_frame_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args(cmd)


@pytest.mark.parametrize(
    'cmd,expected',
    [
        (['--plot', 'title=mytitle'], {'title': 'mytitle'}),
        (
            ['--plot', 'title=mytitle', 'colors=a,b,c'],
            {
                'title': 'mytitle',
                'colors': ('a', 'b', 'c')
                }),
        ],
    )
def test_plot_arg(cmd, expected):
    """Test plot args."""
    parser = argparse.ArgumentParser()
    lc.add_plot_arg(parser)
    v = vars(parser.parse_args(cmd))
    assert v['plot'] is True
    assert 'plotvars' in v
    assert v['plotvars'] == expected


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-o', '_frame0.pdb'),
        ('', False),
        ('--top-output', '_frame0.pdb'),
        ('--top-output frame0_', 'frame0_'),
        ],
    )
def test_topoutput_args(cmd, expected):
    """Test top output."""
    parser = argparse.ArgumentParser()
    lc.add_top_output_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['top_output'] == expected
    

@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-d traj.pdb', 'traj.pdb'),
        ('', 'traj_out.dcd'),
        ('--traj-output traj_out.pdb', 'traj_out.pdb'),
        ('--traj-output newtraj.xtc', 'newtraj.xtc'),
        ],
    )
def test_trajoutput_args(cmd, expected):
    """Test traj output."""
    parser = argparse.ArgumentParser()
    lc.add_traj_output_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['traj_output'] == expected


@pytest.mark.parametrize(
    'cmd',
    [
        ('-t'),
        ('--traj-output'),
        ],
    )
def test_traj_output_arg_error(cmd):
    """Test top output error."""
    parser = argparse.ArgumentParser()
    lc.add_traj_output_arg(parser)
    with pytest.raises(SystemExit):
        parser.parse_args([cmd])


@pytest.mark.parametrize(
    'cmd,expected',
    [
        ('-x', 'results.csv'),
        ('--export', 'results.csv'),
        ('-x data.csv', 'data.csv'),
        ('--export data.csv', 'data.csv'),
        ('', False),
        ],
    )
def test_data_export_arg(cmd, expected):
    """Test data export arg."""
    parser = argparse.ArgumentParser()
    lc.add_data_export_arg(parser)
    v = vars(parser.parse_args(cmd.split()))
    assert v['export'] == expected



"""Contains plotting routines.

Plotting routines can be hardcoded or linked from thirdparty projects.

References
----------

* https://python-bioplottemplates.readthedocs.io/en/latest/
"""
from bioplottemplates.plots import plotlabeld, plotparam

import taurenmd.core as tcore
from taurenmd.libs import libcli


@libcli.add_reference(tcore.ref_plottemplates_param)
def param(*args, **kwargs):
    """
    Reproduce bioplotemplates.plots.param.
    
    Visit: https://python-bioplottemplates.readthedocs.io/en/latest/
    """
    return plotparam(*args, **kwargs)


@libcli.add_reference(tcore.ref_plottemplates_labeldots)
def label_dots(*args, **kwargs):
    """
    Reproduce bioplotemplates.plots.label_dots.

    Visit: https://python-bioplottemplates.readthedocs.io/en/latest/
    """
    return plotlabeld(*args, **kwargs)



"""
Library package.

Taurenmd libraries, all prefixed as ``lib``, contain the functions used
by the client interfaces. Though taurenmd is designed to be used as a
command-line interface, we provide detailed documentation for its
internal functions.

Lib packages are organized by their scope, some have ``libs`` have
general scopes while other are tightly bound to the Molecular Dynamics
library with which they operate. For example:

#. ``libmda`` for MDAnalysis
#. ``libmdt`` for MDTraj

These modules store functions that relate solely to the scope of
MDAnalysis and MDTraj packages; normally these are I/O related operations.

On the other hand, we decided to organize other libraries based on the
scope of their functions regardless of the dependencies they use.
For example, ``libcalc`` contains functions calculate MD parameters,
and combine the usage of different libraries when needed.

Further instructions are provided within each module documentation.
"""



"""
# Calculates distances between centers of geometry of two selections.

Distance is given in 3D XYZ coordinate space units.

## Algorithm

Distance between centers of geometry is calculated by::

    np.linalg.norm(np.subtract(coord1, coord2))

Where, ``coord*`` are the centers of geometry of each atom selection
``-l1`` and ``-l2``, respectively.
Read further on [np.linalg.norm](https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.norm.html)
and [np.subtract](https://docs.scipy.org/doc/numpy/reference/generated/numpy.subtract.html?highlight=subtract#numpy-subtract).

## Examples

Calculate the distances between two carbon alphas:

    taurenmd dist top.pdb traj.dcd -l1 'resnum 10 and name CA' -l2 'resnum 20 and name CA'

Calculate the distances between two chains:

    taurenmd dist top.pdb traj.dcd -l1 'segid A' -l2 'segid B'

``-x`` exports the data to a CSV file. You can also plot the data with
the ``-v`` option:

    [...] -x distances.csv -v title=my-plot-title xlabel=frames ylabel=degrees ...

where ``[...]`` is the previous command example.

``dist`` can be run directly as main command instead of subroutine:

    tmddist

## References

"""  # noqa: E501
import argparse
import functools

import numpy as np

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log
from taurenmd.libs import libcli, libio, libmda, libplot  # noqa: F401
from taurenmd.logger import S, T


__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_plottemplates_param}'
    )

_help = 'Calculates distances between geometric centers of selections. '
_name = 'dist'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)

ap.add_argument(
    '-l1',
    '--sel1',
    help='First selection.',
    default='all',
    type=str,
    )

ap.add_argument(
    '-l2',
    '--sel2',
    help='Second selection.',
    default='all',
    type=str,
    )

libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        sel1='all',
        sel2='all',
        start=None,
        stop=None,
        step=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs
        ):
    """Execute main client logic."""
    log.info(T('measuring distances'))
    
    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    u = libmda.load_universe(topology, *trajectories)
    
    frame_slice = libio.frame_slice(
        start=start,
        stop=stop,
        step=step,
        )
   
    log.info(T('defining atom seletions'))
    log.info(S('atom selection #1: {}', sel1))
    log.info(S('atom selection #2: {}', sel2))
    atom_sel1 = u.select_atoms(sel1)
    atom_sel2 = u.select_atoms(sel2)

    distances = np.ones(len(u.trajectory[frame_slice]), dtype=np.float32)
    
    log.info(T('Calculating distances'))
    # https://www.mdanalysis.org/MDAnalysisTutorial/atomgroups.html
    # https://www.mdanalysis.org/docs/documentation_pages/core/groups.html#MDAnalysis.core.groups.AtomGroup.center_of_geometry
    for i, _ts in enumerate(u.trajectory[frame_slice]):

        distances[i] = np.linalg.norm(
            np.subtract(
                atom_sel1.center_of_geometry(),
                atom_sel2.center_of_geometry(),
                )
            )
    
    log.info(S('calculated a total of {} distances.', len(distances)))
    
    if export:
        libio.export_data_to_file(
            list(range(len(u.trajectory))[frame_slice]),
            distances,
            fname=export,
            header=(
                '# Distances between two selections centers of geomemtry\n'
                '# topology: {}\n'
                '# trajectories: {}\n'
                '# selection #1: {}\n'
                '# selection #2: {}\n'
                '# frame,distance\n'
                ).format(
                    topology,
                    ', '.join(t.resolve().str() for t in trajectories),
                    sel1,
                    sel2,
                    ),
            )

    if plot:
        plotvars = plotvars or dict()
        plotvars.setdefault('labels', '{} dist {}'.format(sel1, sel2))

        log.info(T('plot params:'))
        for k, v in plotvars.items():
            log.info(S('{} = {!r}', k, v))

        libplot.param(
            list(range(len(u.trajectory))[frame_slice]),
            distances,
            **plotvars,
            )

    log.info(S('done'))
    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
# Decompose Eurler angle rotations of a selection.

**EXPERIMENTAL PROTOCOL, RESULTS MAY NOT BE RELIABLE**

*Calculate the Roll, Pitch and Yaw angles along the trajectory.*

Read further on roll, pitch and yaw angles (Euler Angles) -
[wikipedia](https://en.wikipedia.org/wiki/Euler_angles).

Here we decompose these movements around the three different axis
centered at an origin using Quaternion rotation.

## Algorithm

Given a selection of three regions, ``selection A``, ``selection B``
and ``selection C``:

1. Centers the system to the three selections center of geometry for
every frame, this is called the *origin*,

2. Calculates a plane given by the center of geometries of the three
selections, plane ABC,

3. Defines the vector OA that goes from the origin to the center of
geometry of ``selection A``, this represents the Yaw axis.

4. Defines the normal vector to the plane ABC (ABCn), this represents the
Roll axis,

5. Defines the cross product betwee vectors OA and ABCn (AONn), this is the
Pitch axis,

6. Calculates this axis of reference for every frame

Calculating the angles:

Angles represent the right hand rotation around an axis of the sistem in
a i-frame compared to the reference frame. Quaterion distance is calculated
by [libcalc.generate_quaternion_rotations](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#generate_quaternion_rotations) and
[libcal.sort_by_minimum_Qdistances](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.sort_by_minimum_Qdistances).

### Roll

The roll angle is calculated by rotating the unitary vector OA
around vector ABCn until the Quaternion distance is the minimum
between the vector OAi (in frame) and vector OA in reference frame.

### Pitch

The pitch angle is calculated by rotating the unitary vector ABCn
around vector AONn until the Quaternion distance is the minimum
between the vector ABCni (in frame) and vector ABCn in reference frame.

### Yaw

The pitch angle is calculated by rotating the unitary vector AONn
around vector OA until the Quaternion distance is the minimum
between the vector AONni (in frame) and vector AONn in reference frame.

## Examples

In the case of an homotrimer, define the axis and the origin on the trimers:

    taurenmd rorations -z 'segid A' 'segid B' 'segid C' -x rotations.csv

## References

"""  # noqa: E501
import argparse
import functools

import numpy as np

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log
from taurenmd.libs import libcalc, libcli, libio, libmda  # noqa: F401
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_pyquaternion}'
    )

_help = 'Calculates angular rotations across axes.'
_name = 'rotations'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_plane_selection_arg(ap)
libcli.add_angle_unit_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        plane_selection,
        aunit='degrees',
        start=None,
        stop=None,
        step=None,
        export=False,
        **kwargs,
        ):
    """Execute main client logic."""
    log.info(T('starting'))
    
    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    u = libmda.load_universe(topology, *trajectories)

    log.info(T('transformation'))
    fSlice = libio.frame_slice(start=start, stop=stop, step=step)
    
    origin_selection = ' or '.join(plane_selection)
    pABC_atomG = u.select_atoms(origin_selection)
    ABC_selections = plane_selection
    # p stands for point
    pA_atomG = u.select_atoms(ABC_selections[0])
    pB_atomG = u.select_atoms(ABC_selections[1])
    pC_atomG = u.select_atoms(ABC_selections[2])

    u.trajectory[0]
    
    # defining the center of reference
    pABC_cog = pABC_atomG.center_of_geometry()
    log.info(T('Original Center of Geometry'))
    log.info(S('for frame: 0'))
    log.info(S('for selection: {}', origin_selection))
    log.info(S('pABC_cog: {}', pABC_cog))
    
    log.info(T('Transfering'))
    log.info(S('all coordinates of reference frame to the origin 0, 0, 0'))
    pABC_atomG.positions = pABC_atomG.positions - pABC_cog
    log.info(S('COG in origin: {}', pABC_atomG.center_of_geometry()))

    log.info(T('defining the reference axes'))
    pA_cog = pA_atomG.center_of_geometry()
    pB_cog = pB_atomG.center_of_geometry()
    pC_cog = pC_atomG.center_of_geometry()
    log.info(S('plane points definition:'))
    log.info(S('pA: {}', pA_cog))
    log.info(S('pB: {}', pB_cog))
    log.info(S('pC: {}', pC_cog))

    log.info(T('defining the normal vector to reference plane'))
    ref_plane_normal = libcalc.calc_plane_normal(pA_cog, pB_cog, pC_cog)
    log.info(S('plane normal: {}', ref_plane_normal))
    log.info(S('done'))

    log.info(T('defining the cross product vector'))
    ref_plane_cross = np.cross(pA_cog, ref_plane_normal)
    log.info(S('np cross: {}', ref_plane_cross))
    
    roll_angles = []
    pitch_angles = []
    yaw_angles = []

    for i, _ts in enumerate(u.trajectory[fSlice]):
        print(f'.. working for frame :{i}')

        pABC_cog_ts = pABC_atomG.center_of_geometry()
        pABC_atomG.positions = pABC_atomG.positions - pABC_cog_ts

        pA_cog_ts = pA_atomG.center_of_geometry()
        pB_cog_ts = pB_atomG.center_of_geometry()
        pC_cog_ts = pC_atomG.center_of_geometry()

        ts_plane_normal = libcalc.calc_plane_normal(
            pA_cog_ts,
            pB_cog_ts,
            pC_cog_ts,
            )

        ts_plane_cross = np.cross(pA_cog_ts, ts_plane_normal)

        # Calculating Quaternion Rotations
        roll_Qs_tuples = libcalc.generate_quaternion_rotations(
            ref_plane_normal,
            pA_cog_ts,
            )

        pitch_Qs_tuples = libcalc.generate_quaternion_rotations(
            ref_plane_cross,
            ts_plane_normal,
            )

        yaw_Qs_tuples = libcalc.generate_quaternion_rotations(
            pA_cog,
            ts_plane_cross,
            )

        roll_minimum = libcalc.sort_by_minimum_Qdistances(
            roll_Qs_tuples,
            pA_cog,
            )[0][0]

        pitch_minimum = libcalc.sort_by_minimum_Qdistances(
            pitch_Qs_tuples,
            ref_plane_normal,
            )[0][0]

        yaw_minimum = libcalc.sort_by_minimum_Qdistances(
            yaw_Qs_tuples,
            ref_plane_cross,
            )[0][0]
        
        if aunit == 'degrees':
            roll_angles.append(round(roll_minimum.degrees, 3))
            pitch_angles.append(round(pitch_minimum.degrees, 3))
            yaw_angles.append(round(yaw_minimum.degrees, 3))
        else:
            roll_angles.append(round(roll_minimum.radians, 3))
            pitch_angles.append(round(pitch_minimum.radians, 3))
            yaw_angles.append(round(yaw_minimum.radians, 3))
    
    if export:
        file_names = []
        for _fname in ['roll', 'pitch', 'yaw']:
            file_names.append(
                libio.add_prefix_to_path(
                    export,
                    f"{_fname}_angles_",
                    )
                )

        log.info(T('Saving data to files'))
        for data, fname in zip(
                [roll_angles, pitch_angles, yaw_angles],
                file_names,
                ):

            log.info(S('saving {}', fname))
            libio.export_data_to_file(
                list(range(len(u.trajectory))[fSlice]),
                data,
                fname=fname,
                header=(
                    '# Topology: {}\n'
                    '# Trajectories: {}\n'
                    '# Plane Selection: {}\n'
                    '# frame,ange{}\n'
                    ).format(
                        topology,
                        ', '.join(t.resolve().str() for t in trajectories),
                        origin_selection,
                        aunit,
                        ),
                )
        log.info(S('done'))

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""Test logger module."""

import taurenmd.logger as tl


def test_has_T():
    """Test has T."""
    assert hasattr(tl, 'T')


def test_has_S():
    """Test has S."""
    assert hasattr(tl, 'S')


def test_TitleLog():
    """Test TitleLog formatter."""
    t = tl.TitleLog('section title: {}', 'chapter 1')
    assert '\n* Section Title: chapter 1 ...' == str(t)


def test_SubLog():
    """Test SubLog formatter."""
    s = tl.SubLog('some args: {}, {}', '1', '2')
    assert '    some args: 1, 2' == str(s)


def test_SubLog_2():
    """Test SubLog formatter."""
    s = tl.SubLog('some args: {}, {}', '1', '2', indent=1, spacer='*')
    assert '****some args: 1, 2' == str(s)



"""
Functions that wrap around `MDAnalysis library`_.

Functions contained in this module operate with MDAnalysis (MDA)
functionalities, either by using MDA to access Molecular Dynamics
data or by receiving MDA data structures and parsing them in some way.

When using functions contained in this library you should `cite both`_
taurenmd and MDAnalysis.

.. _MDAnalysis library: https://www.mdanalysis.org
.. _cite both: https://taurenmd.readthedocs.io/en/latest/citing.html
"""
import MDAnalysis as mda
from MDAnalysis.analysis import align as mdaalign

import taurenmd.core as tcore
from taurenmd import Path, log
from taurenmd.libs import libcli, libio
from taurenmd.logger import S, T


@libcli.add_reference(tcore.ref_mda)
def load_universe(topology, *trajectories):
    """
    Load MDAnalysis universe.

    Accepts MDAnalysis compatible `topology formats`_ and
    `trajectory formats`_. Read further on the `MDAnalysis Universe`_.

    .. _topology formats: https://www.mdanalysis.org/docs/documentation_pages/topology/init.html
    .. _trajectory formats: https://www.mdanalysis.org/docs/documentation_pages/coordinates/init.html#supported-coordinate-formats
    .. _MDAnalysis Universe: https://www.mdanalysis.org/docs/documentation_pages/core/universe.html?highlight=universe#core-object-universe-mdanalysis-core-universe

    Examples
    --------
        
        >>> libmda.load_universe('topology.pdb', 'trajectory.dcd')

        >>> libmda.load_universe(
                'topology.pdb',
                'traj_part_1.xtc',
                'traj_part_2.xtc',
                Path('my', 'md', 'folder', 'traj_part_3.xtc'),
                )

    Parameters
    ----------
    topology : str or Path object
        Path to topology file.

    trajectories* : str of Path objects
        Paths to trajectory file(s). Trajectory files will be used
        sequentially to create the Universe.
    
    Return
    ------
    MDAnalysis Universe
    """  # noqa: E501 D412
    libio.report_input(topology, trajectories)
    universe = mda.Universe(
        Path(topology).str(),
        [Path(i).str() for i in trajectories],
        )
    report(universe)
    return universe


@libcli.add_reference(tcore.ref_mda)
def report(universe):
    """
    Report information about the Universe.
    
    Example
    -------

        >>> u = libmda.load_universe('topology.pdb', 'trajectory.xtc')
        >>> libmda.report(u)
       
    Parameters
    ----------
    universe : MDAnalysis Universe
        `MDAnalysis universe <https://www.mdanalysis.org/docs/documentation_pages/core/universe.html?highlight=universe#core-object-universe-mdanalysis-core-universe>`_.

    Returns
    -------
    None

    """  # noqa: E501
    log.info(T('Reporting'))
    log.info(S('number of frames: {}', len(universe.trajectory)))
    log.info(S('number of atoms: {}', len(universe.atoms)))


@libcli.add_reference(tcore.ref_mda)
def mdaalignto(universe, reference, selection='all'):
    """
    Align universe to reference.

    Uses `MDAnalysis.analysis.align.alignto <https://www.mdanalysis.org/docs/documentation_pages/analysis/align.html?highlight=alignto#MDAnalysis.analysis.align.alignto>`_.
    
    Parameters
    ----------
    universe, reference, selection
        Same as in ``MDAnalysis.analysis.align.alignto`` function.

    Raises
    ------
    ZeroDivisionError
        If selection gives empty selection.
    """  # noqa: E501
    try:
        mdaalign.alignto(universe, reference, select=selection)
    except ZeroDivisionError as err:
        log.debug(err, exc_info=True)
        errmsg = (
            f'Could not perform alignment due to {err}, '
            'most likely the alignment selection does not match '
            'any possible selection in the system. You selection : '
            f'\'{selection}\'.'
            )
        log.info(errmsg)
        raise err


@libcli.add_reference(tcore.ref_mda)
def draw_atom_label_from_atom_group(atom_group):
    """
    Translate MDAnalysis Atom Group to list of strings for each atom.

    Strings represent each atom by SEGID.RESNUM|RESNAME.NAME,
    for example carbon alpha of Cys 18 of chain A would:
        
        >>> A.18Cys.CA

    This function is used by taurenmd for data representation
    purporses.

    Parameters
    ----------
    atom_group : Atom Group obj
       `MDAnalysis Atom group <https://www.mdanalysis.org/docs/documentation_pages/core/groups.html?highlight=atom%20group#MDAnalysis.core.groups.AtomGroup>`_.

    Returns
    -------
    list of strings
        Containing the atom string representation for each atom in
        the Atom Group.
    """  # noqa: E501
    labels = []
    for atom in atom_group:
        s = '{}.{}{}.{}'.format(
            atom.segment.segid,
            atom.residue.resnum,
            atom.resname.title(),
            atom.name,
            )
        labels.append(s)

    return labels



"""
# Edit a trajectory.

*In short*, takes a trajectory and apply a modification:

1. frame slicing
2. atom selection
3. unwrap
4. align
5. file format change

Saves the result to a new trajectory.

All operations are performed with `MDAnalysis <https://www.mdanalysis.org>`_.

## Examples

Changes trajectory file format

    taurenmd trajedit top.pdb traj.xtd -d traj.dcd

Extracts a part of the trajectory atoms, in this example ``segid A``,
the option ``-o`` saves the first frame of the new trajectory to a topology
file:

    taurenmd trajedit top.pdb traj.xtc -d tsegidA.dcd -o -l "segid A"

You can slice the trajectory by appending the following ``-s``, ``-e`` or
``-p`` options, this saves only every 100 frames:

    [...] -p 100

You can align the trajectory to a part of the system, for example,
align the whole system to one of its subunits:

    taurenmd trajedit top.pdb traj.dcd -d alignedA.dcd -a "segid A and name CA"

further restrain the output to a specific subselection with ``-l``:
    
    [...] -l "segid A or segid B"

``trajedit`` also implements the ``unwrap`` method from which is an
alternative approach to the ``imagemol`` client, that implements from
``MDTraj``. See references section.

    taurenmd trajedit top.pdb traj.dcd -d unwrapped.dcd -w -o unwrapped_frame0.pdb


## References

"""  # noqa: E501
import argparse
import functools

import MDAnalysis as mda

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, _controlled_exit, log
from taurenmd.libs import libcli, libio, libmda
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_mda_unwrap}'
    f'{tcore.ref_mda_alignto}'
    )

_help = 'Edits trajectory in many different ways.'
_name = 'trajedit'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_insort_arg(ap)
libcli.add_atom_selection_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_traj_output_arg(ap)
libcli.add_top_output_arg(ap)


ap.add_argument(
    '-a',
    '--align',
    help=(
        'Align system to a atom group. '
        'Alignmed RMSD is compared to the Topology coordinates. '
        'Uses MDAnalysis.analysis.align.alignto. '
        'If given without argument defaults to \'all\'. '
        'Defaults to ``False``.'
        ),
    default=False,
    const='all',
    nargs='?',
    )

ap.add_argument(
    '-w',
    '--unwrap',
    help=(
        'Unwraps selection according to: '
        'https://www.mdanalysis.org/docs/documentation_pages/core/groups.html#MDAnalysis.core.groups.AtomGroup.unwrap'  # noqa: E501
        ),
    action='store_true',
    )

ap.add_argument(
    '--unwrap-reference',
    help=(
        'The unwrap method reference parameter. '
        'Has effect only if \'-w\' is given. '
        'Defaults to `None`.'
        ),
    default=None,
    )

ap.add_argument(
    '--unwrap-compound',
    help=(
        'The unwrap method compound parameter. '
        'Has effect only if \'-w\' is given. '
        'Defaults to `fragments`.'
        ),
    default='fragments',
    type=str,
    )


def _ap():
    return ap


def main(
        topology,
        trajectories,
        insort=None,
        start=None,
        stop=None,
        step=None,
        selection='all',
        traj_output='traj_out.dcd',
        top_output=None,
        unwrap=False,
        unwrap_reference=None,
        unwrap_compound='fragments',
        align=False,
        **kwargs,
        ):
    """Execute main client logic."""
    log.info(T('editing trajectory'))
    
    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    if insort:
        trajectories = libio.sort_numbered_input(*trajectories)

    u = libmda.load_universe(topology, *trajectories)
    
    if unwrap:
        log.info(T('unwrapping'))
        log.info(S('set to: {}', unwrap))
        log.info(S('reference: {}', unwrap_reference))
        log.info(S('compound: {}', unwrap_compound))

    if align:
        log.info(T('Alignment'))
        log.info(S('trajectory selection will be aligned to subselection:'))
        log.info(S('- {}', align, indent=2))
        align_reference = mda.Universe(Path(topology).str())
    
    log.info(T('transformation'))
    sliceObj = libio.frame_slice(start, stop, step)

    log.info(S('selecting: {}', selection))
    atom_selection = u.select_atoms(selection)
    log.info(S('with {} atoms', atom_selection.n_atoms, indent=2))

    log.info(T('saving trajectory'))
    traj_output = Path(traj_output)
    log.info(S('destination: {}', traj_output.resolve().str()))

    with mda.Writer(traj_output.str(), atom_selection.n_atoms) as W:
        for i, _ts in zip(
                range(len(u.trajectory))[sliceObj],
                u.trajectory[sliceObj],
                ):
            
            log.info(S('working on frame: {}', i))
            
            if unwrap:
                log.debug(S('unwrapping', indent=2))
                atom_selection.unwrap(
                    reference=unwrap_reference,
                    compound=unwrap_compound,
                    )

            if align:
                try:
                    libmda.mdaalignto(u, align_reference, selection=align)
                except ZeroDivisionError:
                    _controlled_exit()

            W.write(atom_selection)
    
    log.info(S('trajectory saved'))

    if top_output:
        log.info(T('saving topology'))
        fout = libio.parse_top_output(top_output, traj_output)
        log.info(S('saving frame 0 to: {}', fout.resolve()))
        with mda.Writer(fout.str(), atom_selection.n_atoms) as W:
            for _ts in u.trajectory[sliceObj][0:1]:
                if unwrap:
                    log.debug(S('unwrapping for topology', indent=2))
                    atom_selection.unwrap(
                        reference=unwrap_reference,
                        compound=unwrap_compound,
                        )
                W.write(atom_selection)
    
    log.info(S('Done'))
    return


maincli = functools.partial(libcli.maincli, ap, main)

if __name__ == '__main__':
    maincli()



"""
# Report on trajectory characteristics.

## Example

     taurenmd report topology.pdb trajectory.dcd

## References

"""
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import _BANNER, log
from taurenmd.libs import libcli, libmda
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += tcore.ref_mda

_help = 'Reports on trajectory details.'
_name = 'report'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)


def _ap():
    return ap


def main(topology, trajectories, **kwargs):
    """Execute main client logic."""
    log.info(T('reporting'))
    libmda.load_universe(topology, *trajectories)
    log.info(S('done'))
    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
# Extract trajectory frames to individual files.

Normally used to extract frames to PDB topology files so those can
be inspected independently.

## Note

    Frame number is 0-indexed.

## Examples:

Extract frames 11 to 49 (inclusive), remember frames index start at 0:

    taurenmd fext topology.pdb trajectory.dcd -s 10 -e 50

Extract the first frame:

    taurenmd fext topology.pdb trajectory.dcd -flist 0

Extract a selection of frames:

    taurenmd fext topology.pdb trajectory.dcd -flist 0,10,23,345

Frame file types can be specified:

    taurenmd fext topology.pdb trajectory.dcd -p 10 -x .dcd

  
Atom selection can be specified as well, the following extracts
only the 'segid A' atom region of the first frame. Selection rules
are as decribed for [MDAnalysis selection](https://www.mdanalysis.org/docs/documentation_pages/selections.html).

    taurenmd fext topology.pdb trajectory.xtc -flist 0 -l 'segid A'

Multiple trajectories can be given, they will be contatenated:

    taurenmd fext top.pdb traj1.xtc traj2.xtc traj3.xtc -p 10

Can also be used as main command:

    tmdfext topology.pdb ...


## References:

"""  # noqa: E501
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log
from taurenmd.libs import libcli, libio, libmda
from taurenmd.logger import S


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    )

_help = 'Extract trajectory frames to individual files.'
_name = 'fext'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_atom_selection_arg(ap)
libcli.add_frame_list_arg(ap)
libcli.add_slice_arg(ap)

ap.add_argument(
    '-f',
    '--prefix',
    help='String prefix for each file. Defaults to `frame_`.',
    default='frame_',
    )

ap.add_argument(
    '-x',
    '--ext',
    help='Extension of frame files. Defaulst to .pdb',
    default='.pdb',
    )


def _ap():
    return ap


def main(
        topology,
        trajectories,
        start=None,
        stop=None,
        step=None,
        flist=None,
        prefix='frame_',
        ext='pdb',
        selection='all',
        **kwargs):
    """Execute main client logic."""
    log.info('Starting...')
    
    u = libmda.load_universe(topology, *trajectories)
    
    frames_to_extract = libio.frame_list(
        len(u.trajectory),
        start=start,
        stop=stop,
        step=step,
        flist=flist,
        )

    log.info(S('extracting {} frames', len(frames_to_extract)))
    
    zeros = len(str(len(u.trajectory)))
    ext = ext.lstrip('.').strip()
    
    atom_group = u.select_atoms(selection)

    for frame in frames_to_extract:
        file_name = '{}{}.{}'.format(
            prefix,
            str(frame).zfill(zeros),
            ext,
            )

        atom_group.write(
            filename=Path(file_name),
            frames=[frame],
            )

        log.info(S('writen frame {}, to {}', frame, file_name))

    return


maincli = functools.partial(libcli.maincli, ap, main)

if __name__ == '__main__':
    maincli()



"""Test libplot."""

from taurenmd import Path
from taurenmd.libs import libplot as lplt


def test_param():
    """Test param plot."""
    p = Path('param.pdf')
    lplt.param(
        list(range(10)),
        list(range(10)),
        filename='param.pdf',
        )
    assert p.exists()
    p.unlink()


def test_label_dots():
    """Test label_dots plot."""
    p = Path('label_dots.pdf')
    lplt.param(
        ['a', 'b', 'c'],
        list(range(3)),
        filename='label_dots.pdf',
        )
    assert p.exists()
    p.unlink()



"""
taurenmd main client interface.

Usage Examples
--------------

To access to the complete list of *taurenmd commands* with a summary
information for each, execute:

    >>> taurenmd -h

or simply:

    >>> taurenmd

To see the current version number:

    >>> taurenmd -v

Using ``trajedit`` as an example, lets inspect its functionality:

    >>> taurenmd trajedit -h

With ``trajedit`` you can edit a trajectory in many different ways.
For example, convert a trajectory to another format:

    >>> taurenmd trajedit topology.pdb trajectory.xtc -d new_trajectory.dcd

The above command reads the original ``trajectory.xtc`` file and outputs
the new ``new_trajectory.dcd``. You can also use ``trajedit`` to reduce
the trajectory size, say by slicing every 10 frames:

    >>> taurenmd trajedit topology.pdb trajetory.xtc -d traj_p10.xtc -p 10

the ``-p`` option refers to the slicing step size, in this case ``10`` -
reads every 10 frames. Likewise, you can pass a *start* (``-s``) and an
*end* (``-e``) arguments:

    >>> taurenmd trajedit topology.pdb trajectory.xtc -d traj_s50_e500_p10.xtc -s 50 -e 500 -p 10

Also, you can extract an Atom Selection from a trajectory to a new trajectory
file. The example bellow creates a new trajectory from the input one containing
only atoms belonging to chain A. In cases like this it is useful to extract
the atom selection as an independent topology file.

    >>> taurenmd trajedit top.pdb traj.xtc -l 'segid A' -d chainA.xtc -o chainA_topology.pdb

You can also use ``trajedit`` to extract a specific frame from a trajectory:

    >>> taurenmd trajedit topology.pdb trajectory.xtc -d frame40.pdb -s 40 -e 41

but, for this example, you could instead use the ``fext`` interface:

    >>> taurenmd fext topology.pdb trajectory.xtc -f 40 -x .pdb -f frame_

Each an every ``taurenmd`` sub command is available directly as a main
routine by prefixing a ``tmd`` to its name, for example:

    >>> taurenmd trajedit
    >>> # equals to
    >>> tmdtrajedit

Logging
-------

*taurenmd* logs all its running activity as follows:

1. ``.taurenmd.cmd``, keeps an historic register of the taurenmd commands
run on a given folder together with a list of the research projects
that must be cited for that particular run; these are the libraries taurenmd
used to access and process the MD data. It also servers as a record for your
research project.

2. ``.taurenmd.log``, a user readable logging information, the very same that is
printed in the ``terminal`` during runtime. Overwrites previous runs.

3. ``.taurenmd.debug``, a full verbose log with all runtime information for the
LAST run. Overwrites previous runs.
"""  # noqa: E501
import argparse
import sys

# add bellow, by alphabetical order your newly developed cli
# import taurenmd.cli_NAME as cli_NAME
import taurenmd.cli_distances as cli_dist
import taurenmd.cli_fext as cli_fext
import taurenmd.cli_imagemol as cli_imagemol
import taurenmd.cli_nosol as cli_nosol
import taurenmd.cli_pangle as cli_pangle
import taurenmd.cli_report as cli_report
import taurenmd.cli_rmsd as cli_rmsd
import taurenmd.cli_rmsf as cli_rmsf
import taurenmd.cli_rotations as cli_rot
import taurenmd.cli_trajedit as cli_trajedit
from taurenmd import _INTERFACE_DESCRIPTION, log
from taurenmd.libs import libcli
from taurenmd.logger import CMDFILE


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
# add yourself to the credits
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

ap = libcli.CustomParser(
    description=_INTERFACE_DESCRIPTION + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)

subparsers = ap.add_subparsers(
    title='taurenmd subroutines',
    )

# add your client to this block following the example of the
# other clients.
# libcli.add_subparser(subparsers, cli_NAME)
libcli.add_subparser(subparsers, cli_pangle)
libcli.add_subparser(subparsers, cli_dist)
libcli.add_subparser(subparsers, cli_fext)
libcli.add_subparser(subparsers, cli_imagemol)
libcli.add_subparser(subparsers, cli_nosol)
libcli.add_subparser(subparsers, cli_report)
libcli.add_subparser(subparsers, cli_rmsd)
libcli.add_subparser(subparsers, cli_rmsf)
libcli.add_subparser(subparsers, cli_rot)
libcli.add_subparser(subparsers, cli_trajedit)

# Done :-) There is nothing else to do here if you are implementing a new
# client interface.


def _ap():
    return ap


def load_args():
    """Load user input arguments."""
    return ap.parse_args()


def maincli():
    """Execute main client logic."""
    if len(sys.argv) < 2:
        ap.error('A subcommand is required.')

    args = load_args()
    log.debug(args)
    libcli.save_command(CMDFILE, *sys.argv)
    result = args.func(**vars(args))
    libcli.save_references()
    return result


if __name__ == '__main__':
    maincli()



"""Test core definitions."""
import taurenmd.core as tcore


def test_references():
    """Test scientific references to third party libraries."""
    assert isinstance(tcore.ref_mdt, str)
    assert isinstance(tcore.ref_mda, str)
    assert isinstance(tcore.ref_mda_selection, str)
    assert isinstance(tcore.ref_mda_unwrap, str)
    assert isinstance(tcore.ref_mda_alignto, str)
    assert isinstance(tcore.ref_plottemplates_param, str)
    assert isinstance(tcore.ref_plottemplates_labeldots, str)
    assert isinstance(tcore.ref_pyquaternion, str)
    assert isinstance(tcore.ref_numpy, str)



"""
Template for command-line clients.

This file provides a template structure to build new command line clients.

Copy this file to a new file named for your command line, for example:

    >>> cp _cli_template.py cli_NAME.py

Where ``NAME`` is the name of your new client, do not use CAPS.

In your new client, replace this docstring by a proper description
of the operations performed by your client. Keep clients as simple as possible.
Two simple clients are better than a complex one. You can/should read
through the documentation of other clients for examples.

Code blocks are mandatory by default unless stated as optional.
Also, functions and variable names in the main namespace should be kept
not to break functionality with the program infrastructure.

For those optional code blocks, just delete or comment out those options
needed or not for your purpose.

Please refer to our library documentation to see all available functions.
Feel free to implement you own functions in their respective lib*.py modules.

Once it is finished, add your client to the cli.py file, instructions are
provided there as comments throughout the file.

Also, add an entry point to the setup.py file, instructions are provided
there as well, look for the 'entry_points' argument.

Test driven development is a must. You should write the test cases for
your client and newly developed functions. Visit the tests/test_clis.py,
instructions are provided by comments throughout the script. Add additional
tests on the test_lib* corresponding file if you have developed new functions.

Finally, go the ``docs/reference`` folder. There is a ``cli_template.xxx``
template file there, just copy it to a ``cli_name.rst`` file and modify
it accordingly: 1) give a title, 2) change ``NAME`` for the name of your
client. Add that file reference to the toctree in the
``docs/reference/clients.rst`` file.

You can use ``tox`` during your development to confirm code style,
documentation build, and tests. It is useful if the tests for these
three tox environments pass before you Pull Request:

    >>> tox -e check
    >>> tox -e docs
    >>> tox -e py37  # or py36 depending
"""
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import log
# add here additional libraries if you need them
# libmda, libmdt, libcalc, etc...
from taurenmd.libs import libcli, libio  # noqa: F401
from taurenmd.logger import S, T


__author__ = 'Your Name'
__email__ = 'your_email@address.com'
__maintainer__ = 'Your Name'  # if you plan to give long term maintenance
# otherwise just write, 'taurenmd'.
__credits__ = ['Your Name']
__status__ = 'Production'

# This block adds to the docstring the references used by your client.
# for example, MDAnalysis or MDTraj. Remember to add a:
# **References:**
# line to your client docstring.
# taurenmd.core module contains all strings related to citation.
# please refer to that module. Bellow is an example for referencing
# MDAnalysis.
__doc__ += (
    f'{tcore.ref_mda}'
    # just keep adding as your need
    )

# updated this variables according to your client
_help = 'Short description sentence.'
_name = 'name'  # the actual command subroutine name
# this will be the name to be called with
# $ taurenmd name
# choose wisely, has to sound :-)

# this block initiates the argparse.ArgumentParser
# libcli.CustomParser is just an extension to the argparse.AP
# that provides enhanced help message. Additional information is available
# in the documentation or the libcli.py file itself.
ap = libcli.CustomParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

# libcli contains several functions that add common arguments to the
# argument parser, for example, output trajectory or trajectory slice
# related arguments, so you can use those functions to add those commonly
# used arguments, instead of writing them from scratch.
# Just refer to libcli for more information, those functions
# start with 'add_'.
libcli.add_topology_arg(ap)

# choose between one of the following, whether your client will/could
# receive several trajectories as input or only one. If your client
# is based on MDAnalysis you want the first choice.
# libcli.add_trajectories_arg(ap)
# libcli.add_trajectory_arg(ap)

# but, if you need to add client specific arguments, use the space bellow:
# here you just follow Python argparse interface:
#
# https://docs.python.org/3/library/argparse.html
#
# ap.add_argument(
# ...


# this is used just for documentation purposes.
def _ap():
    return ap


def main(
        topology,
        trajectory,
        # trajectories,
        # additional args...
        # ...
        **kwargs
        ):
    """Execute main client logic."""
    log.info(T('starting'))
    
    # write your logic here.
    # use the log.info and log.debug registries to log and output
    # information to the user.
    # you can always refer to other clients for examples.

    log.info(S('done'))
    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""Test libmda."""
import copy

import MDAnalysis as mda
import numpy as np
import pytest

from taurenmd.libs import libmda as la

from . import toptest, trajtest


def test_load_universe_1():
    """Test load MDA Universe."""
    universe = la.load_universe(
        toptest,
        trajtest,
        )
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 10


def test_load_universe_2():
    """Test load MDA Universe multiple trajs."""
    universe = la.load_universe(
        toptest,
        trajtest,
        trajtest,
        trajtest,
        )
    assert isinstance(universe, mda.Universe)
    assert len(universe.trajectory) == 30

   
def test_report_universe():
    """
    Test universe report.

    Only interface, does not test return value.
    """
    universe = mda.Universe(
        toptest.str(),
        trajtest.str(),
        )
    la.report(universe)


def test_mdaalignto():
    """Test MDA alignto."""
    universe = mda.Universe(toptest.str(), trajtest.str())
    reference = mda.Universe(toptest.str())
    ugroup = universe.select_atoms('all')
     
    for _ts in universe.trajectory[1:]:
        prev = copy.copy(ugroup.positions)
        assert np.all(np.equal(prev, ugroup.positions))
        la.mdaalignto(universe, reference, 'name CA')
        assert np.all(np.not_equal(prev, ugroup.positions))


def test_mdaalignto_exit():
    """Test exit on ZeroDivision."""
    universe = mda.Universe(toptest.str(), trajtest.str())
    reference = mda.Universe(toptest.str())
    with pytest.raises(ZeroDivisionError):
        la.mdaalignto(universe, reference, 'name ZZZZZ')


def test_draw_atom_label():
    """Test label maker."""
    universe = mda.Universe(
        toptest.str(),
        trajtest.str(),
        )
    
    atom_group = universe.select_atoms('resnum 10-12 and name CA')

    labels = la.draw_atom_label_from_atom_group(atom_group)
    assert isinstance(labels, list)
    assert labels == ['A.10Ser.CA', 'A.11Ile.CA', 'A.12Leu.CA']



"""
# Make molecules whole.

Attempts to "Recenter and apply periodic boundary conditions to the
molecules in each frame of the trajectory." (MDTraj documentation)

## Algorithm

Uses [MDTraj.Trajectory.image_molecule](http://mdtraj.org/1.9.3/api/generated/mdtraj.Trajectory.html?highlight=image%20molecules#mdtraj.Trajectory.image_molecules) and
[MDTraj.Topology.find_molecules](http://mdtraj.org/1.9.3/api/generated/mdtraj.Topology.html?highlight=find_molecules#mdtraj.Topology.find_molecules).

### Protocol 1

Performs ``mdtraj.top.find_molecules`` and ``mdtraj.traj.image_molecules``
in the trajectory as a whole. ``anchor_molecules`` parameters
get ``mdtraj.top.find_molecules[:1]``, and ``other_molecules``
parameter receives ``mdtraj.top.find_molecules[1:]``.

### Protocol 2
    
The same as protocol 1 but executes those steps for each frame
separately. Frames are concatenated back to a whole trajectory
at the end.

## Examples

Basic usage, ``-o`` saves the first frame in a separate topology file:

    taurenmd imagemol top.pdb traj.dcd -d imaged.dcd -o

For trajectories with *non-standard* molecules you can use a TPR file.

    taurenmd imagemol top.tpr traj.xtc -d imaged.xtc

Using protocol 2

    taurenmd imagemol top.tpr traj.xtc -d imaged.xtc -i 2

## References

"""  # noqa: E501
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import _BANNER, log
from taurenmd.libs import libcli, libio, libmdt
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mdt}'
    )

_help = 'Attempt to image molecules.'
_name = 'imagemol'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectory_arg(ap)
libcli.add_traj_output_arg(ap)
libcli.add_top_output_arg(ap)

ap.add_argument(
    '-i',
    '--protocol',
    help=(
        'The protocol with which reimage. '
        'Read main command description for details.'
        ),
    default=1,
    type=int,
    )


def _ap():
    return ap


def main(
        topology,
        trajectory,
        traj_output='imaged.dcd',
        top_output=False,
        protocol=1,
        **kwargs
        ):
    """Execute main client logic."""
    log.info('Attempting image molecules')
    
    traj = libmdt.load_traj(topology, trajectory)
    
    protocols = {
        1: libmdt.imagemol_protocol1,
        2: libmdt.imagemol_protocol2,
        }

    reimaged = protocols[protocol](traj)

    log.info(T('saving the output'))
    reimaged.save(traj_output)
    log.info(S('saved trajectory: {}', traj_output))

    if top_output:
        fout = libio.parse_top_output(top_output, traj_output)
        reimaged[0].save(fout.str())
        log.info(S('saving frame 0 to: {}', fout.resolve()))

    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
# Remove solvent from trajectory.

## Algorithm

Removes solvent from trajectory using [MDTraj.remove_solvent](http://mdtraj.org/1.9.3/api/generated/mdtraj.Trajectory.html?highlight=remove%20solvent#mdtraj.Trajectory.remove_solvent).

## Examples

Remove all solvent:

    taurenmd nosol top.pdb traj.dcd -d traj_nosol.dcd -o

Remove all solvent except for NA atoms:

    taurenmd nosol top.pdb traj.dcd -d traj_nosolNA.dcd -e Na -o

``tmdnosol`` can be used as main command:

    tmdnosol [...]

## References

"""  # noqa: E501
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import _BANNER, log
from taurenmd.libs import libcli, libio, libmdt
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += tcore.ref_mdt

_help = 'Removes solvent from trajectory.'
_name = 'nosol'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectory_arg(ap)
libcli.add_traj_output_arg(ap)
libcli.add_top_output_arg(ap)

ap.add_argument(
    '-m',
    '--maintain',
    help=(
        'List of solvent residue names to maintain in trajectory. '
        'Feeds MDTraj.Trajectory.remove_solvent.exclude parameter.'
        ),
    default=None,
    nargs='+',
    )


def _ap():
    return ap


def main(
        topology,
        trajectory,
        selection=None,
        maintain=None,
        top_output='_frame0.pdb',
        traj_output='nosol.dcd',
        **kwargs
        ):
    """Execute main client logic."""
    log.info(T('Removing solvent'))

    trj = libmdt.load_traj(topology, trajectory)
    
    if selection:
        log.info(S('selecting {}', selection))
        atom_sel = trj.top.select(selection)
        trj.atom_slice(atom_sel, inplace=True)

    trj.remove_solvent(inplace=True, exclude=maintain)
    
    if top_output:
        fout = libio.parse_top_output(top_output, traj_output)
        trj[0].save(fout.str())
        log.info(S('first frame topology saved: {}', fout))
   
    log.info(T('saving trajectory'))
    log.info(S('destination: {}', traj_output))
    trj.save(traj_output)
    log.info(S('saved'))


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
# Calculate angular oscillation of a plane along the trajectory.

A plane is defined by the centers of geometry of three atom selection
groups. The angle between that plane in each frame and itself in the
reference frame is computed. Angle can be reported in degrees (default)
or radians.

## Algorithm

Plane equation is computed by [libcalc.calc_plane_eq](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.calc_plane_eq).
Angle between planes is computed by [libcalc.calc_planes_angle](https://taurenmd.readthedocs.io/en/latest/reference/libcalc.html#taurenmd.libs.libcalc.calc_planes_angle).
Refer to our documentation page for more details.

## Examples

Given a protein of 3 subunits (chains or segids) calculate the
angle variation of a plane that crosses the protein longitudinally:

    taurenmd pangle top.pdb traj.xtc -z 'segid A' 'segid B' 'segid C' -x

``-x`` exports the data to a CSV file. You can also plot the data with
the ``-v`` option:

    [...] -v title=my-plot-title xlabel=frames ylabel=degrees ...

where ``[...]`` is the previous command example.

``pangle`` can be run directly as main command instead of subroutine:

    tmdpangle

## References

"""  # noqa: E501
import argparse
import functools

import taurenmd.core as tcore
from taurenmd import _BANNER, Path, log
from taurenmd.libs import libcalc, libcli, libio, libmda, libplot
from taurenmd.logger import S, T


__author__ = 'Joao M.C. Teixeira'
__email__ = 'joaomcteixeira@gmail.com'
__maintainer__ = 'Joao M.C. Teixeira'
__credits__ = ['Joao M.C. Teixeira']
__status__ = 'Production'

__doc__ += (
    f'{tcore.ref_mda}'
    f'{tcore.ref_mda_selection}'
    f'{tcore.ref_plottemplates_param}'
    )

_help = 'Calculate angular oscillation of a plane along the trajectory.'
_name = 'pangle'

ap = libcli.CustomParser(
    description=_BANNER + __doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )

libcli.add_version_arg(ap)
libcli.add_topology_arg(ap)
libcli.add_trajectories_arg(ap)
libcli.add_plane_selection_arg(ap)
libcli.add_angle_unit_arg(ap)
libcli.add_reference_frame_arg(ap)
libcli.add_slice_arg(ap)
libcli.add_data_export_arg(ap)
libcli.add_plot_arg(ap)


def _ap():
    return ap


def main(
        topology,
        trajectories,
        plane_selection,
        aunit='degrees',
        ref_frame=0,
        start=None,
        stop=None,
        step=None,
        export=False,
        plot=False,
        plotvars=None,
        **kwargs
        ):
    """Execute main client logic."""
    log.info(T('calculating angles'))
    
    topology = Path(topology)
    trajectories = [Path(t) for t in trajectories]

    u = libmda.load_universe(topology, *trajectories)

    frame_slice = libio.frame_slice(
        start=start,
        stop=stop,
        step=step,
        )
    log.info(S('for slice {}', frame_slice))

    log.info(T('calculating plane eq. for reference frame'))
    log.info(S('using frame: {}', ref_frame))
    u.trajectory[ref_frame]
    reference_point_1 = u.select_atoms(plane_selection[0]).center_of_geometry()
    reference_point_2 = u.select_atoms(plane_selection[1]).center_of_geometry()
    reference_point_3 = u.select_atoms(plane_selection[2]).center_of_geometry()

    ra, rb, rc, rd = libcalc.calc_plane_eq(
        reference_point_1,
        reference_point_2,
        reference_point_3,
        )
    log.info(S('the equation is {}x + {}y + {}z = {}', ra, rb, rc, rd))
    
    log.info(T('Calculating angles'))
    angles = []
    for _ts in u.trajectory[frame_slice]:

        point1 = u.select_atoms(plane_selection[0]).center_of_geometry()
        point2 = u.select_atoms(plane_selection[1]).center_of_geometry()
        point3 = u.select_atoms(plane_selection[2]).center_of_geometry()
        
        a, b, c, d = libcalc.calc_plane_eq(point1, point2, point3)

        angles.append(
            libcalc.calc_planes_angle(
                ra, rb, rc, a, b, c,
                aunit=aunit,
                )
            )
    
    log.info(S('calculated a total of {} angles.', len(angles)))
   
    if export:
        libio.export_data_to_file(
            list(range(len(u.trajectory))[frame_slice]),
            angles,
            fname=export,
            header=(
                '# Angular oscillation between a plane representatives\n'
                '# topology: {}\n'
                '# trajectories: {}\n'
                '# selections: {}\n'
                '# frame,angle({})\n'
                ).format(
                    topology,
                    ', '.join(t.resolve().str() for t in trajectories),
                    plane_selection,
                    aunit,
                    ),
            )

    if plot:
        plotvars = plotvars or dict()
        plotvars.setdefault(
            'labels',
            'plane: {}'.format(' and '.join(plane_selection)),
            )

        log.info(T('plot params:'))
        for k, v in plotvars.items():
            log.info(S('{} = {!r}', k, v))

        libplot.param(
            list(range(len(u.trajectory))[frame_slice]),
            angles,
            **plotvars,
            )

    log.info(S('done'))
    return


maincli = functools.partial(libcli.maincli, ap, main)


if __name__ == '__main__':
    maincli()



"""
Functions that wrap around `MDTraj library`_.

Functions contained in this module operate with MDTraj functionalities,
wither by using MDTraj to access Molecular Dynamics data or by receiving
MDTraj data structures and parsing them in some way.

`Simtk OpenMM`_ is also used in some functions.

Read our `citing documentation`_ to understand how to cite multiple
libraries.

.. _MDTraj library: http://mdtraj.org/1.9.3/index.html
.. _citing documentation: https://taurenmd.readthedocs.io/en/latest/citing.html
.. _Simtk OpenMM: http://openmm.org/
"""
import sys

import mdtraj

import taurenmd.core as tcore
from taurenmd import Path, log
from taurenmd.libs import libcli, libio
from taurenmd.logger import S, T


try:
    import simtk.openmm.app as app
    SIMTK = True
except ImportError:
    SIMTK = False


def _log_simtkimport_error():
    msg = (
        "To use .cif files as Topologies taurenmd requires OpenMM, "
        "which is currently not installed. "
        "Please visit our Installation instruction at "
        "https://taurenmd.readthedocs.io/"
        )
    log.error(T('Dependency Error'))
    log.error(S(msg))
    sys.exit(0)


@libcli.add_reference(tcore.ref_mdt)
def load_traj(topology, trajectory):
    """
    Load trajectory with `MDTraj <http://mdtraj.org/1.9.3/index.html>`_.
    
    Uses `mdtraj.load <http://mdtraj.org/1.9.3/api/generated/mdtraj.load.html?highlight=load#mdtraj.load>`_.
    
    Example
    -------
        
        >>> libmdt.load_traj('bigtopology.cif', 'trajectory.dcd')

    Parameters
    ----------
    topology : str or Path
        Path to the topology file. Accepts MDTraj compatible `topology files <http://mdtraj.org/1.9.3/load_functions.html#trajectory-reference>`_. mmCIF format is loaded using `OpenMM <http://mdtraj.org/1.9.3/api/generated/mdtraj.Topology.html?highlight=from_openmm#mdtraj.Topology.from_openmm>`_.

    trajectory : str or Path
        Path to the trajectory file. Accepts MDTraj compatible `files <http://mdtraj.org/1.9.3/load_functions.html#trajectory-reference>`_

    Returns
    -------
    MDTraj trajectory
        `Trajectory object <http://mdtraj.org/1.9.3/api/generated/mdtraj.Trajectory.html#mdtraj-trajectory>`_.
    """  # noqa: E501
    libio.report_input(topology, trajectory)

    topp = Path(topology)
    if topp.suffix == '.cif' and SIMTK:
        mol = app.PDBxFile(topp.str())
        top = mdtraj.Topology.from_openmm(mol.topology)
    elif topp.suffix == '.cif' and not SIMTK:
        _log_simtkimport_error()
    else:
        top = topp.str()

    mdtrajectory = mdtraj.load(Path(trajectory).str(), top=top)

    return mdtrajectory


@libcli.add_reference(tcore.ref_mdt)
def imagemol_protocol1(traj):
    """Attempt to image molecules acting on the whole traj."""
    log.info(T('running reimage protocol #1'))
    log.info(S('finding molecules'))

    mols = traj.top.find_molecules()
    log.info(S('done'))
    
    log.info(T('reimaging'))
    reimaged = traj.image_molecules(
        inplace=False,
        anchor_molecules=mols[:1],
        other_molecules=mols[1:],
        )
    log.info(S('done'))
    return reimaged


@libcli.add_reference(tcore.ref_mdt)
def imagemol_protocol2(traj):
    """Attempt to image molecules frame by frame."""
    reimaged = []
    for frame in range(len(traj)):
        log.info(S('reimaging frame: {}', frame))
        
        mols = traj[frame].top.find_molecules()
    
        reimaged.append(
            traj[frame].image_molecules(
                inplace=False,
                anchor_molecules=mols[:1],
                other_molecules=mols[1:],
                )
            )

    log.info(S('concatenating traj frames'))
    # http://mdtraj.org/1.9.3/api/generated/mdtraj.join.html#mdtraj.join
    reimaged_traj = reimaged[0].join(reimaged[1:])

    return reimaged_traj



"""
Installation setup.

Read the full installation instructions at:

    https://taurenmd.readthedocs.io/en/latest/installation.html

To install taurenmd run:

    >>> python setup.py

To install taurenmd in developer mode run:

    >>> python setup.py develop

To install taurenmd without ANY of its dependencies:

    >>> python setup.py --no-deps
"""
from __future__ import absolute_import, print_function

import io
import os
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


install_requires = [
    'bioplottemplates>0.1',
    'pyquaternion',
    ]

if os.getenv('READTHEDOCS'):
    install_requires = []

# support deps
supdeps = [
    'matplotlib>=3,<4',
    'numpy>=1,<2',
    ]

# Molecular Dynamics deps
mddeps = [
    'MDAnalysis>=0.2,<1',
    'mdtraj>=1,<2',
    ]

alldeps = mddeps + supdeps


def _read(*names, **kwargs):
    with io.open(
            join(dirname(__file__), *names),
            encoding=kwargs.get('encoding', 'utf8')
            ) as fh:
        return fh.read()


setup(
    name='taurenmd',
    version='0.8.10',
    license='GNU GPLv2',
    description='A command-line interface for analysis routines in Molecular Dynamics data.',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', _read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', _read('CHANGELOG.rst'))
        ),
    author='Joao MC Teixeira',
    author_email='joaomcteixeira@gmail.com',
    url='https://github.com/joaomcteixeira/taurenmd',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities',
        ],
    project_urls={
        'Documentation': 'https://taurenmd.readthedocs.io/',
        'Changelog': 'https://taurenmd.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/joaomcteixeira/taurenmd/issues',
        },
    keywords=[
        'Molecular Dynamics',
        'Proteins',
        'DNA',
        'RNA',
        'Structural Biology',
        'Molecular Biology',
        'Biochemistry',
        ],
    python_requires='>=3.6, <3.8',
    install_requires=install_requires,
    extras_require={
        'all': alldeps,
        'md': mddeps,
        'sup': supdeps,
        },
    entry_points={
        'console_scripts': [
            'taurenmd = taurenmd.cli:maincli',
            # add bellow, by alphabetical, order your newly developed
            # client, following the nomenclature provided,
            # where NAME is the command line name of your subroutine.
            # 'tmdNAME = taurenmd.cli_NAME:maincli',
            'tmdpangle = taurenmd.cli_pangle:maincli',
            'tmddist = taurenmd.cli_distances:maincli',
            'tmdfext = taurenmd.cli_fext:maincli',
            'tmdimagemol = taurenmd.cli_imagemol:maincli',
            'tmdnosol = tauremd.cli_nosol:maincli',
            'tmdreport = taurenmd.cli_report:maincli',
            'tmdrot = taurenmd.cli_rot:maincli',
            'tmdrmsf = taurenmd.cli_rmsf:maincli',
            'tmdrmsd = taurenmd.cli_rmsd:maincli',
            'tmdtrajedit = taurenmd.cli_trajedit:maincli',
            ]
        },
    )



"""Manages operations with logging."""

DEBUGFILE = '.taurenmd.debug'
LOGFILE = '.taurenmd.log'
CMDFILE = '.taurenmd.cmd'


class TitleLog:
    """Format string to title."""

    def __init__(self, msg, *args):
        self.msg = msg.title()
        self.args = args
    
    def __str__(self):
        output = self.msg.format(*self.args)
        return '\n* {} ...'.format(output)


class SubLog:
    """
    Format string to bullet point like structure.
    
    This format performs nicely under the `TitleLog` formatting.
    """
    
    def __init__(self, msg, *args, indent=1, spacer=' '):
        self.msg = msg
        self.args = args
        self.indent = spacer * 4 * indent
    
    def __str__(self):
        output = self.msg.format(*self.args)
        return '{}{}'.format(self.indent, output)


T = TitleLog
S = SubLog



# THIS FILE IS EXCLUDED FROM FLAKE8 IN TOX.ini WORKFLOW
r"""
Welcome to
```
_________ _______           _______  _______  _        _______  ______  
\__   __/(  ___  )|\     /|(  ____ )(  ____ \( (    /|(       )(  __  \ 
   ) (   | (   ) || )   ( || (    )|| (    \/|  \  ( || () () || (  \  )
   | |   | (___) || |   | || (____)|| (__    |   \ | || || || || |   ) |
   | |   |  ___  || |   | ||     __)|  __)   | (\ \) || |(_)| || |   | |
   | |   | (   ) || |   | || (\ (   | (      | | \   || |   | || |   ) |
   | |   | )   ( || (___) || ) \ \__| (____/\| )  \  || )   ( || (__/  )
   )_(   |/     \|(_______)|/   \__/(_______/|/    )_)|/     \|(______/ 
                                                                       
```
**A command-line interface for Molecular Dynamics Analysis routines.**

version: {}
"""
# link to logo
# http://patorjk.com/software/taag/#p=display&h=0&f=Epic&t=taurenmd
import logging
import sys

from taurenmd.core import CONTACTUS, Path  # noqa: F401
from taurenmd.logger import DEBUGFILE, LOGFILE

__version__ = '0.8.10'

_BANNER = __doc__.format(__version__)

_DOCUMENTATION = """
Documentation
=============
This project is fully documented at: https://taurenmd.readthedocs.io/
Client documentation is provided with the ``-h`` option, for example:

    >>> taurenmd -h

    or

    >>> taurenmd dist -h

Citing
======
taurenmd uses other scientific libraries to handle and generate
Molecular Dynamics data. You SHOULD by all means cite also the other
libraries when using taurenmd. Please visit our documentation page
for complete details on how to cite properly:

    https://taurenmd.readthedocs.io/en/latest/citing.html
""".format(__version__)

__doc__ = _BANNER + _DOCUMENTATION

_INTERFACE_DESCRIPTION = __doc__ + "Usage\n=====\n\n"

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

_db = logging.FileHandler(DEBUGFILE, mode='w')
_db.setLevel(logging.DEBUG)
_db.setFormatter(
    logging.Formatter(
        '%(filename)s:%(name)s:%(funcName)s:%(lineno)d: %(message)s'
        )
    )

_info = logging.FileHandler(LOGFILE, mode='w')
_info.setLevel(logging.INFO)
_info.setFormatter(logging.Formatter('%(message)s'))

_ch = logging.StreamHandler()
_ch.setLevel(logging.INFO)
_ch.setFormatter(logging.Formatter('%(message)s'))

log.addHandler(_db)
log.addHandler(_info)
log.addHandler(_ch)

references = set()


def _controlled_exit():
    log.info(core.controlled_exit)
    log.info(core.CONTACTUS)
    # https://shapeshed.com/unix-exit-codes/
    # http://www.tldp.org/LDP/abs/html/exitcodes.html
    sys.exit(126)



"""Test libio."""
import pytest

from taurenmd.core import Path
from taurenmd.libs import libio as io

from . import datap, export_data_expected, export_data_expected_2


@pytest.mark.parametrize(
    'in1,expected',
    [
        ('traj_1.dcd', 1),
        ('traj2.dcd', 2),
        ('tr4j2.dcd', 2),
        ('traj_3.dcd', 3),
        ('traj_1231.dcd', 1231),
        ('traj_0011.dcd', 11),
        ('traj_1_.dcd', 1),
        ('traj_20200101_1.dcd', 1),
        ],
    )
def test_get_number(in1, expected):
    """Test get number from path."""
    result = io.get_number(in1)
    assert result == expected


@pytest.mark.parametrize(
    'in1,expected',
    [
        (
            ['t_1.dcd', 't_11.dcd', 't_2.dcd'],
            ['t_1.dcd', 't_2.dcd', 't_11.dcd'],
            ),
        (['b.dcd', 'c.dcd', 'a.dcd'], ['a.dcd', 'b.dcd', 'c.dcd']),
        ],
    )
def test_sort_numbered_input_1(in1, expected):
    """Test sort numbered inputs."""
    result = io.sort_numbered_input(*in1)
    assert result == expected


@pytest.mark.parametrize(
    'in1,error',
    [
        (['t_1.dcd', 't_11.dcd', 't_2.dcd'], TypeError),
        ]
    )
def test_sort_numbered_inputs_error(in1, error):
    """Test sort numbered inputs raised Errors."""
    with pytest.raises(error):
        io.sort_numbered_input(in1)


def test_report_input():
    """
    Test report_input.

    Tests only the interface, there is no return value.
    """
    io.report_input('topology.pdb', 'traj1.dcd')


@pytest.mark.parametrize(
    'ipath,prefix,expected',
    [
        ('traj_output.xtc', 'my_prefix', 'my_prefixtraj_output.xtc'),
        ('traj_output.xtc', 'my_prefix_', 'my_prefix_traj_output.xtc'),
        (None, 'my_prefix', 'my_prefix'),
        (None, 'my_prefix.pdb', 'my_prefix.pdb'),
        ],
    )
def test_add_prefix_to_path(ipath, prefix, expected):
    """Test adding prefix to path."""
    result = io.add_prefix_to_path(ipath, prefix)
    assert isinstance(result, Path)
    assert result.name == expected


@pytest.mark.parametrize(
    'ipath,suffix,expected',
    [
        ('traj_output.xtc', 'my_suffix', 'traj_outputmy_suffix.xtc'),
        ('traj_output.xtc', '_my_suffix', 'traj_output_my_suffix.xtc'),
        ('traj_output.xtc', '_my_suffix.pdb', 'traj_output_my_suffix.pdb'),
        ('traj_output', '_my_suffix', 'traj_output_my_suffix'),
        (None, '_my_suffix', '_my_suffix'),
        (None, '_my_suffix.pdb', '_my_suffix.pdb'),
        ],
    )
def test_add_suffix_to_path(ipath, suffix, expected):
    """Test adding prefix to path."""
    result = io.add_suffix_to_path(ipath, suffix)
    assert isinstance(result, Path)
    assert result.name == expected


@pytest.mark.parametrize(
    'inputs,expected',
    [
        (('traj',), 'traj_frame0.pdb'),
        (('traj', 1), 'traj_frame1.pdb'),
        (('traj', 401), 'traj_frame401.pdb'),
        (('traj', 4, 'dcd'), 'traj_frame4.dcd'),
        (('traj', 4, 'dcd', 4), 'traj_frame0004.dcd'),
        (('traj', 4, 'dcd', 4, '_suffixed.tpr'), 'traj_suffixed.tpr'),
        ],
    )
def test_mk_frame_path(inputs, expected):
    """
    Test mk frame path.

    Notice ``inputs`` is a tuple to unpack.
    """
    result = io.mk_frame_path(*inputs)
    assert isinstance(result, Path)
    assert result.name == expected


@pytest.mark.parametrize(
    'in1,expected',
    [
        ('file', Path('file')),
        (None, Path()),
        ],
    )
def test_get_ipath(in1, expected):
    """Test get ipath private function."""
    result = io._get_ipath(in1)
    assert isinstance(result, Path)
    assert result == expected


@pytest.mark.parametrize(
    'top,traj,expected',
    [
        ('top.pdb', 'traj.xtc', 'top.pdb'),
        (Path('top.pdb'), 'traj.xtc', 'top.pdb'),
        ('_top.pdb', 'traj.xtc', 'traj_top.pdb'),
        ('_top', 'traj.xtc', 'traj_top.xtc'),
        ('top_', 'traj.xtc', 'top_traj.xtc'),
        ],
    )
def test_parse_top_output(top, traj, expected):
    """Test parse topology output."""
    result = io.parse_top_output(top, traj_output=traj)
    assert isinstance(result, Path)
    assert result.name == expected


def test_export_data_to_file_1():
    """Test export data to file."""
    xdata = list(range(10))
    ydata = list(range(0, 20, 2))
    header = "#this is an header\n"
    fout = Path(datap, 'exporttest.csv')
    io.export_data_to_file(
        xdata,
        ydata,
        fname=fout,
        header=header,
        delimiter='-',
        fmt='{:.1f}',
        )

    fres = fout.open().read()
    fexp = export_data_expected.open().read()
    assert fres == fexp
    fout.unlink()


def test_export_data_to_file_2():
    """Test export data to file several series."""
    xdata = list(range(10))
    ydata = list(range(0, 20, 2))
    ydata = [ydata, ydata]
    assert isinstance(ydata, list)
    assert len(ydata) == 2
    header = "#this is an header\n"
    fout = Path(datap, 'exporttest.csv')
    io.export_data_to_file(
        xdata,
        *ydata,
        fname=fout,
        header=header,
        delimiter='-',
        fmt='{:.1f}',
        )

    fres = fout.open().read()
    fexp = export_data_expected_2.open().read()
    assert fres == fexp
    fout.unlink()


@pytest.mark.parametrize(
    'inputs,expected',
    [
        ((10,), list(range(10))),
        ((10, 2, 5), list(range(2, 5))),
        ((10, 2, 50), list(range(2, 10))),
        ((10, None, None, 2), list(range(0, 10, 2))),
        ((10, None, None, None, '1,2,45,65'), [1, 2, 45, 65]),
        ((10, None, None, None, [1, 2, 45, 65]), [1, 2, 45, 65]),
        ((10, None, None, None, ['1', '2', '45', '65']), [1, 2, 45, 65]),
        ((None, None, None, None, ['1', '2', '45', '65']), [1, 2, 45, 65]),
        ],
    )
def test_frame_list(inputs, expected):
    """Test make frame list."""
    result = io.frame_list(*inputs)
    assert result == expected


@pytest.mark.parametrize(
    'lent,flist,error',
    [
        (10, 123, ValueError),
        (None, None, TypeError),
        ]
    )
def test_frame_list_error(lent, flist, error):
    """Test frame list error."""
    with pytest.raises(error):
        io.frame_list(lent, flist=flist)


def test_frame_slice():
    """Test frame_slice."""
    result = io.frame_slice(1, 100, 2)
    assert result == slice(1, 100, 2)


@pytest.mark.parametrize(
    'value,start,stop,step,expected',
    [
        ('1,100,2', None, None, None, slice(1, 100, 2)),
        ('::2', None, None, None, slice(None, None, 2)),
        ('1:100:2', None, None, None, slice(1, 100, 2)),
        ('1:None:2', None, None, None, slice(1, None, 2)),
        (None, 10, None, None, slice(10, None, None)),
        (slice(None), None, None, None, slice(None, None, None)),
        (None, 10, 100, None, slice(10, 100, None)),
        (None, 10, 100, 3, slice(10, 100, 3)),
        ((0, 50, 3), None, None, None, slice(0, 50, 3)),
        ((0, '50', 3), None, None, None, slice(0, 50, 3)),
        ((None, 100, None), None, None, None, slice(None, 100, None)),
        ('10', None, None, None, slice(10, None, None)),
        (10, None, None, None, slice(None, 10, None)),
        (None, None, None, None, slice(None, None, None)),
        (None, {}, None, None, slice(None, None, None)),
        ],
    )
def test_evaluate_to_slice(value, start, stop, step, expected):
    """Test evaluate to slice."""
    result = io.evaluate_to_slice(
        value=value,
        start=start,
        stop=stop,
        step=step,
        )
    assert result == expected


def test_evaluate_to_slice_error():
    """Test evaluate to slice ValueError."""
    with pytest.raises(ValueError):
        io.evaluate_to_slice(value={})
